# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter <gabriel[]potter[]fr>

"""
LDAP Hero: a LDAP browser based on the Scapy LDAP client
"""

import uuid

from scapy.layers.ldap import (
    LDAP_AttributeValue,
    LDAP_BIND_MECHS,
    LDAP_Client,
    LDAP_CONTROL_ACCESS_RIGHTS,
    LDAP_Control,
    LDAP_DS_ACCESS_RIGHTS,
    LDAP_Exception,
    LDAP_ModifyRequestChange,
    LDAP_PartialAttribute,
    LDAP_PROPERTY_SET,
    LDAP_serverSDFlagsControl,
)
from scapy.layers.dcerpc import (
    DCERPC_Transport,
    NDRUnion,
    DCE_C_AUTHN_LEVEL,
    find_dcerpc_interface,
)
from scapy.layers.gssapi import SSP
from scapy.layers.msrpce.rpcclient import (
    DCERPC_Client,
)
from scapy.layers.msrpce.msdrsr import (
    DRS_EXTENSIONS_INT,
    DRS_EXTENSIONS,
    DRS_MSG_CRACKREQ_V1,
    IDL_DRSBind_Request,
    IDL_DRSCrackNames_Request,
    NTDSAPI_CLIENT_GUID,
)
from scapy.layers.ntlm import NTLMSSP
from scapy.layers.kerberos import KerberosSSP
from scapy.layers.spnego import SPNEGOSSP
from scapy.layers.smb2 import (
    SECURITY_DESCRIPTOR,
    WELL_KNOWN_SIDS,
    WINNT_ACE_FLAGS,
    WINNT_ACE_HEADER,
    WINNT_SID,
    WINNT_ACCESS_ALLOWED_ACE,
    WINNT_ACCESS_ALLOWED_OBJECT_ACE,
    WINNT_ACCESS_DENIED_OBJECT_ACE,
    WINNT_ACCESS_DENIED_ACE,
    WINNT_SYSTEM_AUDIT_OBJECT_ACE,
    WINNT_SYSTEM_AUDIT_ACE,
)
from scapy.utils import valid_ip

try:
    import tkinter as tk
    from tkinter import ttk, messagebox
except ImportError:
    raise ImportError("tkinter is not installed (`apt install python3-tk` on debian)")


class AutoHideScrollbar(ttk.Scrollbar):
    def __init__(self, *args, **kwargs):
        self.shown = False
        super(AutoHideScrollbar, self).__init__(*args, **kwargs)

    def set(self, first, last):
        show = float(first) > 0 or float(last) < 1
        if show and not self.shown:
            self.grid(row=0, column=1, sticky="nsew")
        elif not show and self.shown:
            self.grid_forget()
        self.shown = show
        super(AutoHideScrollbar, self).set(first, last)


class BasePopup:
    """
    A tkinter wrapper used to have a popup window with basic controls
    """

    def __init__(self, parent):
        # Get dialog
        self.dlg = tk.Toplevel(parent)
        self.parent = parent
        self.cancelled = False

        # Configure some bindings
        self.dlg.bind("<Return>", self.dismiss)
        self.dlg.bind("<KP_Enter>", self.dismiss)

    def dismiss(self, *_) -> None:
        """
        Close the popup
        """
        self.dlg.grab_release()
        self.dlg.destroy()

    def cancel(self) -> None:
        """
        Cancel the popup
        """
        self.cancelled = True
        self.dismiss()

    def run(self) -> False:
        """
        Show the popup. Returns True if cancelled, False otherwise.
        """
        self.dlg.protocol("WM_DELETE_WINDOW", self.dismiss)
        self.dlg.transient(self.parent)
        self.dlg.wait_visibility()
        self.dlg.grab_set()
        self.dlg.wait_window()

        return self.cancelled


class LDAPHero:
    """
    LDAP Hero - LDAP GUI browser over Scapy's LDAP_Client

    :param ssp: the SSP object to use when binding.
    :param mech: the LDAP_BIND_MECHS to use when binding.
    :param simple_username: if provided, used for Simple binding (instead of the 'ssp')
    :param simple_password:
    :param encrypt: request encryption by default (useful when using 'ssp')
    :param host: auto-connect to a specific host
    :param port: the port to connect to (default: 389/636)
                 (This is only in use when using 'host')
    :param ssl: whether to use SSL to connect or not
                (This is only in use when using 'host')
    """

    def __init__(
        self,
        ssp: SSP = None,
        mech: LDAP_BIND_MECHS = None,
        simple_username: str = None,
        simple_password: str = None,
        encrypt: bool = False,
        host: str = None,
        port: int = None,
        ssl: bool = False,
    ):
        self.client = LDAP_Client()
        self.ssp = ssp
        self.mech = mech
        self.simple_username = simple_username
        self.simple_password = simple_password
        self.encrypt = encrypt
        # Session parameters
        self.connected = False
        self.bound = False
        self.host = host
        self.port = port
        self.ssl = ssl
        self.dns_domain_name = ""
        self.rootDSE = {}
        self.sids = dict(WELL_KNOWN_SIDS)
        self.sidscombo = {}
        self.guids = {}
        self.guidscombo = {"None": None}
        self.guidscomboobject = {"None": None}
        self.loadedSchemaIDGuids = False
        self.crop_output = None
        self.currently_editing = None
        # UI cache
        self.lastSearchString = ""
        # Launch
        self.main()

    def connect(self):
        """
        Connect command.
        """
        # If host is None, we need to ask for it via a dialog.
        if self.host is None:
            # Get dialog
            popup = BasePopup(self.root)
            dlg = popup.dlg

            # Connect UI
            serverv = tk.StringVar()
            serverv.set(self.host or "")
            ttk.Label(dlg, text="Server").grid(row=0, column=0)
            serverf = tk.Entry(dlg, textvariable=serverv)
            serverf.grid(row=0, column=1)

            portv = tk.StringVar()
            portv.set("389")
            ttk.Label(dlg, text="Port").grid(row=1, column=0)
            tk.Entry(dlg, textvariable=portv).grid(row=1, column=1)

            sslv = tk.BooleanVar()
            ttk.Label(dlg, text="SSL").grid(row=2, column=0)
            ttk.Checkbutton(dlg, variable=sslv).grid(row=2, column=1)

            ttk.Button(dlg, text="OK", command=popup.dismiss).grid(row=3, column=0)
            ttk.Button(dlg, text="Cancel", command=popup.cancel).grid(row=3, column=1)

            serverf.focus()

            # Setup
            if popup.run():
                # Cancelled
                return

            # Get values
            self.host = serverv.get()
            try:
                self.port = int(portv.get())
            except ValueError:
                return
            self.ssl = sslv.get()

        # Connect now !
        self.tprint(
            "client.connect(host='%s', port=%s, ssl=%s)"
            % (self.host, self.port, self.ssl)
        )
        try:
            self.client.connect(self.host, port=self.port, use_ssl=self.ssl)
        except Exception as ex:
            self.tprint(str(ex))
            raise
        self.tprint("Established connection to %s." % self.host)
        self.connected = True

        # Alright, change the UI.
        self.menu_connection.entryconfig("Connect", state=tk.DISABLED)
        self.menu_connection.entryconfig("Bind", state=tk.ACTIVE)
        self.menu_connection.entryconfig("Disconnect", state=tk.ACTIVE)
        self.menu_browse.entryconfig("Add child", state=tk.ACTIVE)
        self.menu_browse.entryconfig("Modify", state=tk.ACTIVE)
        self.menu_browse.entryconfig("Modify DN", state=tk.ACTIVE)
        self.menu_browse.entryconfig("Search", state=tk.ACTIVE)
        self.menu_view.entryconfig("Tree", state=tk.ACTIVE)

        # Get rootDSE
        self.tprint("Retrieving base DSA information...")
        try:
            results = self.client.search(
                baseObject="",
                scope=0,
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        attrs = results.get("", None)  # root
        if attrs is None:
            return

        self.rootDSE = attrs

        # Get some infos on the server
        try:
            self.dns_domain_name = self.rootDSE["ldapServiceName"][0].split(":")[0]
        except KeyError:
            pass

        # Display
        self._showsearchresult("", results)

        # If we have a SSP, auto-bind.
        if self.ssp is not None:
            self.bind()

    def disconnect(self):
        """
        Disconnect command.
        """
        if not self.connected:
            return

        self.tprint("client.close()")
        self.client.close()
        self.connected = False

        self.menu_connection.entryconfig("Connect", state=tk.ACTIVE)
        self.menu_connection.entryconfig("Bind", state=tk.DISABLED)
        self.menu_connection.entryconfig("Disconnect", state=tk.DISABLED)
        self.menu_browse.entryconfig("Add child", state=tk.DISABLED)
        self.menu_browse.entryconfig("Modify", state=tk.DISABLED)
        self.menu_browse.entryconfig("Modify DN", state=tk.DISABLED)
        self.menu_browse.entryconfig("Search", state=tk.DISABLED)
        self.menu_view.entryconfig("Tree", state=tk.DISABLED)

    def bind(self, *args):
        """
        Bind command.
        """
        if not self.connected:
            return

        if self.bound:
            # We are re-binding !
            self.ssp = None
            self.bound = False

        if self.ssp is not None or self.simple_username is not None:
            # We have an SSP. Don't prompt
            self.tprint("client.bind(%s, ssl=self.ssp)" % self.mech)
            try:
                self.client.bind(
                    self.mech,
                    ssp=self.ssp,
                    simple_username=self.simple_username,
                    simple_password=self.simple_password,
                    encrypt=self.encrypt,
                )
            except LDAP_Exception as ex:
                self.tprint(
                    ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                    tags=["error"],
                )
                return
            except Exception as ex:
                self.tprint(str(ex))
                raise
            self.tprint("Authenticated.\n", tags=["bold"])
            self.bound = True
            return

        # Get dialog
        popup = BasePopup(self.root)
        dlg = popup.dlg

        # Bind UI
        userv = tk.StringVar()
        ttk.Label(dlg, text="User").grid(row=0, column=0)
        userf = tk.Entry(dlg, textvariable=userv)
        userf.grid(row=0, column=1)

        passwordv = tk.StringVar()
        ttk.Label(dlg, text="Password").grid(row=1, column=0)
        tk.Entry(dlg, textvariable=passwordv).grid(row=1, column=1)

        domainv = tk.StringVar()
        domainv.set(self.dns_domain_name)
        ttk.Label(dlg, text="Domain").grid(row=2, column=0)
        domentry = tk.Entry(dlg, textvariable=domainv)
        domentry.grid(row=2, column=1)

        bindtypefrm = ttk.LabelFrame(
            dlg,
            text="Bind type",
        )
        bindtypev = tk.StringVar()
        ttk.Radiobutton(
            bindtypefrm,
            variable=bindtypev,
            text="Sicily bind (NTLM)",
            value=LDAP_BIND_MECHS.SICILY.value,
        ).pack(anchor=tk.W)
        ttk.Radiobutton(
            bindtypefrm,
            variable=bindtypev,
            text="GSSAPI bind (Kerberos)",
            value=LDAP_BIND_MECHS.SASL_GSSAPI.value,
        ).pack(anchor=tk.W)
        ttk.Radiobutton(
            bindtypefrm,
            variable=bindtypev,
            text="SPNEGO bind (NTLM/Kerberos)",
            value=LDAP_BIND_MECHS.SASL_GSS_SPNEGO.value,
        ).pack(anchor=tk.W)
        ttk.Radiobutton(
            bindtypefrm,
            variable=bindtypev,
            text="Simple bind",
            value=LDAP_BIND_MECHS.SIMPLE.value,
        ).pack(anchor=tk.W)
        bindtypefrm.grid(row=3, column=0, columnspan=2)

        encryptv = tk.BooleanVar()
        encryptv.set(self.encrypt)
        ttk.Label(dlg, text="Encrypt traffic after bind").grid(row=4, column=0)
        encrbtn = ttk.Checkbutton(dlg, variable=encryptv)
        encrbtn.grid(row=4, column=1)

        ttk.Button(dlg, text="OK", command=popup.dismiss).grid(row=5, column=0)
        ttk.Button(dlg, text="Cancel", command=popup.cancel).grid(row=5, column=1)

        # Default state
        if self.dns_domain_name and not valid_ip(self.host):
            bindtypev.set(LDAP_BIND_MECHS.SASL_GSS_SPNEGO.value)
        else:
            domentry.configure(state=tk.DISABLED)
            bindtypev.set(LDAP_BIND_MECHS.SICILY.value)

        # Handle dynamic UI
        def bindtypechange(*args, **kwargs):
            bindtype = LDAP_BIND_MECHS(bindtypev.get())
            if bindtype == LDAP_BIND_MECHS.SIMPLE:
                domentry.config(state=tk.DISABLED)
                encrbtn.config(state=tk.DISABLED)
                encryptv.set(False)
            elif bindtype == LDAP_BIND_MECHS.SICILY:
                domentry.config(state=tk.DISABLED)
                encrbtn.config(state=tk.NORMAL)
            else:
                domentry.config(state=tk.NORMAL, textvariable=domainv)
                encrbtn.config(state=tk.NORMAL)

        bindtypev.trace("w", bindtypechange)
        userf.focus()

        # Setup
        if popup.run():
            # Cancelled
            return

        # Get values
        username = userv.get()
        password = passwordv.get()
        domain = domainv.get()
        bindtype = LDAP_BIND_MECHS(bindtypev.get())
        encrypt = encryptv.get()

        # Bind !
        self.tprint("client.bind(%s, ...)" % bindtype)
        try:
            simple_username = None
            simple_password = None
            if bindtype == LDAP_BIND_MECHS.SIMPLE:
                self.ssp = None
                simple_username = username
                simple_password = password
                encrypt = False
            elif bindtype == LDAP_BIND_MECHS.SICILY:
                self.ssp = NTLMSSP(
                    UPN=username,
                    PASSWORD=password,
                )
            elif bindtype == LDAP_BIND_MECHS.SASL_GSSAPI:
                self.ssp = KerberosSSP(
                    UPN="%s@%s" % (username, domain),
                    SPN="ldap/%s" % self.host,
                    PASSWORD=password,
                )
            elif bindtype == LDAP_BIND_MECHS.SASL_GSS_SPNEGO:
                self.ssp = SPNEGOSSP(
                    [
                        NTLMSSP(
                            UPN=username,
                            PASSWORD=password,
                        ),
                        KerberosSSP(
                            UPN="%s@%s" % (username, domain),
                            SPN="ldap/%s" % self.host,
                            PASSWORD=password,
                        ),
                    ]
                )
            self.client.bind(
                bindtype,
                ssp=self.ssp,
                simple_username=simple_username,
                simple_password=simple_password,
                encrypt=encrypt,
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            # Reset SSP.
            self.ssp = None
            return
        except Exception as ex:
            self.tprint(str(ex))
            # Reset SSP.
            self.ssp = None
            raise
        self.tprint("Authenticated.\n")
        self.bound = True

    def tree(self, *args):
        """
        Tree command.
        """
        if not self.connected:
            return

        # Get namingContexts from rootDSE
        try:
            results = self.client.search(attributes=["namingContexts"])
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        attrs = results.get("", None)  # root
        if attrs is None:
            return

        if "namingContexts" in attrs:
            self.tk_tree.delete(*self.tk_tree.get_children())
            for root in attrs["namingContexts"]:
                self.tk_tree.insert("", "end", root, text=root)

    def _showsearchresult(self, baseObject, results):
        """
        Display attributes search result
        """
        if baseObject in results:
            self.tprint("Dn: %s" % (baseObject or "(RootDSE)"), tags=["bold"])
            self.tprint(
                "\n".join(
                    "    %s%s: %s"
                    % (
                        k,
                        "" if len(v) == 1 else " (%s)" % len(v),
                        self._format_attribute(k, v, crop=True),
                    )
                    for k, v in sorted(results[baseObject].items(), key=lambda x: x[0])
                )
                + "\n"
            )

    def treedoubleclick(self, _):
        """
        Action done on tree double-click.
        """
        # Get clicked item
        item = self.tk_tree.selection()[0]

        # Unclickable
        if self.tk_tree.tag_has("unclickable", item):
            return

        # Does it already have children? If so delete them.
        self.tk_tree.delete(*self.tk_tree.get_children(item))

        self.tprint("-----------\nExpanding base '%s'..." % item)

        # Get children
        try:
            results = self.client.search(
                baseObject=item,
                scope=1,
                attributes=["1.1"],
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        # Add to tree
        if not results:
            self.tk_tree.insert(item, "end", text="No children", tags=("unclickable",))
        else:
            for child in results:
                self.tk_tree.insert(item, "end", child, text=child)

        # Get attributes
        try:
            results = self.client.search(
                baseObject=item,
                scope=0,
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        # Display
        self._showsearchresult(item, results)

    def load_guids(self):
        """
        Load the various guids:
        - schemaIDguid
        - propset

        This cache is used to resolve the GUIDs of objects in ACEs.
        """
        if self.loadedSchemaIDGuids:
            return True

        # Property set
        self.guids.update(
            (
                k,
                {
                    "objectClass": ["propset"],
                    "name": v,
                },
            )
            for k, v in LDAP_PROPERTY_SET.items()
        )

        # Control access
        self.guids.update(
            (
                k,
                {
                    "objectClass": ["controlset access right"],
                    "name": v,
                },
            )
            for k, v in LDAP_CONTROL_ACCESS_RIGHTS.items()
        )

        self.tprint("Resolving schemaIDguid... ", flush=True)
        try:
            results = self.client.search(
                baseObject=self.rootDSE["schemaNamingContext"][0],
                scope=1,
                attributes=["lDAPDisplayName", "schemaIDGUID", "objectClass"],
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return False

        self.guids.update(
            {
                uuid.UUID(bytes_le=v["schemaIDGUID"][0]): {
                    "objectClass": v["objectClass"],
                    "name": v["lDAPDisplayName"][0],
                }
                for v in results.values()
                if "schemaIDGUID" in v
            }
        )

        self.guidscombo.update({v["name"]: k for k, v in self.guids.items()})
        self.guidscomboobject.update(
            {
                v["name"]: k
                for k, v in self.guids.items()
                if "classSchema" in v["objectClass"]
            }
        )
        self.loadedSchemaIDGuids = True
        self.tprint("OK !")
        return True

    def _rslvtype(self, x):
        """
        Resolve Object types GUIDs
        """
        if x in self.guids:
            return self.guids[x]["name"]
        return str(x)

    def _rslvsid(self, x):
        """
        Resolve SIDs
        """
        if isinstance(x, WINNT_SID):
            x = x.summary()
        if x in self.sids:
            return self.sids[x]
        return x or ""

    def resolvesids(self, sids):
        """
        Queue a list of SIDs for resolution.
        They are then added to self.sids if successful.
        """
        unknowns = [x for x in (y.summary() for y in sids) if x not in self.sids]
        if not unknowns:
            return

        # Perform a resolution using [MS-LSAT] LsarLookupSids3
        client = DCERPC_Client(
            DCERPC_Transport.NCACN_IP_TCP,
            ndr64=False,
            auth_level=DCE_C_AUTHN_LEVEL.PKT_PRIVACY,
            ssp=self.ssp,
        )
        client.connect_and_bind(self.host, find_dcerpc_interface("drsuapi"))

        # 1. DRSBind
        bind_resp = client.sr1_req(
            IDL_DRSBind_Request(
                puuidClientDsa=NTDSAPI_CLIENT_GUID,
                pextClient=DRS_EXTENSIONS(rgb=bytes(DRS_EXTENSIONS_INT(Pid=1234))),
                ndr64=client.ndr64,
            ),
        )
        if bind_resp.status != 0:
            self.tprint("Bind Request failed.")
            bind_resp.show()
            return

        # 2. DRSCrackNames
        resp = client.sr1_req(
            IDL_DRSCrackNames_Request(
                hDrs=bind_resp.phDrs,
                dwInVersion=1,
                pmsgIn=NDRUnion(
                    tag=1,
                    value=DRS_MSG_CRACKREQ_V1(
                        CodePage=0x4E4,  #
                        LocaleId=0x409,  # US-EN
                        formatOffered=11,  # SID
                        formatDesired=0xFFFFFFF2,  # DS_USER_PRINCIPAL_NAME_FOR_LOGON
                        rpNames=unknowns,
                    ),
                ),
                ndr64=client.ndr64,
            ),
        )
        if resp.status != 0:
            self.tprint("DsCracknames Request failed.")
            resp.show()
            return

        # 3. parse results
        for i, res in enumerate(resp.valueof("pmsgOut.pResult.rItems")):
            if res.status != 0:
                # Errored
                continue
            name = res.valueof("pName")
            self.sids[unknowns[i]] = name.decode()

        # alias for combobox
        self.sidscombo = {self._rslvsid(x): x for x in self.sids.keys()}

    def viewsec(self, *args):
        """
        View security descriptor
        """
        # Get clicked item
        item = self.tk_tree.selection()[0]

        # Get SD
        try:
            results = self.client.search(
                baseObject=item,
                scope=0,
                attributes=["ntSecurityDescriptor"],
                controls=[
                    LDAP_Control(
                        controlType="1.2.840.113556.1.4.801",
                        criticality=True,
                        controlValue=LDAP_serverSDFlagsControl(
                            flags="OWNER+GROUP+DACL+SACL",
                        ),
                    )
                ],
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        if item not in results:
            return

        try:
            nTSecurityDescriptor = SECURITY_DESCRIPTOR(
                results[item]["nTSecurityDescriptor"][0]
            )
        except LDAP_Exception as ex:
            self.tprint(
                "Error parsing the Security Descriptor: " + str(ex),
                tags=["error"],
            )
            return

        # Resolve the guids
        if not self.load_guids():
            return

        # Pre-resolve all the SIDs.
        owner = getattr(nTSecurityDescriptor, "OwnerSid", None)
        group = getattr(nTSecurityDescriptor, "GroupSid", None)
        _to_resolve = [
            owner,
            group,
        ]
        if hasattr(nTSecurityDescriptor, "DACL"):
            _to_resolve.extend(x.Sid for x in nTSecurityDescriptor.DACL.Aces)
        if hasattr(nTSecurityDescriptor, "SACL"):
            _to_resolve.extend(x.Sid for x in nTSecurityDescriptor.SACL.Aces)
        self.resolvesids(_to_resolve)

        # Get dialog
        popup = BasePopup(self.root)
        dlg = popup.dlg

        # Security Descriptor UI
        dlg.columnconfigure(0, weight=1)
        dlg.rowconfigure(tuple(range(5)), weight=1)

        sidfrm = ttk.Frame(dlg)
        sidfrm.grid(row=0, sticky="we")
        sidfrm.grid_columnconfigure(1, weight=1)

        ownerv = tk.StringVar()
        ownerv.set(self._rslvsid(owner))
        ttk.Label(sidfrm, text="Owner").grid(row=0, column=0, sticky="we")
        ttk.Combobox(
            sidfrm, textvariable=ownerv, values=list(self.sidscombo.keys())
        ).grid(row=0, column=1, sticky="we")

        groupv = tk.StringVar()
        groupv.set(self._rslvsid(group))
        ttk.Label(sidfrm, text="Group").grid(row=1, column=0, sticky="we")
        ttk.Combobox(
            sidfrm, textvariable=groupv, values=list(self.sidscombo.keys())
        ).grid(row=1, column=1, sticky="we")

        sdcontrolfrm = ttk.LabelFrame(
            dlg,
            text="SD Control",
        )
        sdflags = [
            "SELF_RELATIVE",
            "DACL_PRESENT",
            "SACL_PRESENT",
            "OWNER_DEFAULTED",
            "DACL_PROTECTED",
            "SACL_PROTECTED",
            "GROUP_DEFAULTED",
            "DACL_AUTO_INHERITED",
            "SACL_AUTO_INHERITED",
            "RM_CONTROL_VALID",
            "DACL_DEFAULTED",
            "SACL_DEFAULTED",
            "SERVER_SECURITY",
            "DACL_COMPUTED",
            "SACL_COMPUTED",
            None,
            "DACL_TRUSTED",
        ]
        sdvars = [None] * len(sdflags)
        for i, sdflag in enumerate(sdflags):
            if sdflag is None:
                continue
            sdvars[i] = tk.BooleanVar()
            sdvars[i].set(getattr(nTSecurityDescriptor.Control, sdflag))
            ttk.Checkbutton(sdcontrolfrm, variable=sdvars[i], text=sdflag).grid(
                row=(i // 3) * 4, column=(i % 3) * 4, columnspan=4, sticky="w"
            )
        sdcontrolfrm.grid(row=1, sticky="we")

        def acegui(ace, parentdlg=dlg):
            data = ace.extractData(accessMask=LDAP_DS_ACCESS_RIGHTS)

            # Sub-dialog
            subpopup = BasePopup(parentdlg)
            dlg = subpopup.dlg

            # Edit ACE UI
            dlg.columnconfigure(1, weight=1)
            dlg.rowconfigure(tuple(range(8)), weight=1)

            # Trustee
            trusteev = tk.StringVar()
            trusteev.set(self._rslvsid(data["sid-string"]))
            ttk.Label(dlg, text="Trustee").grid(row=0, column=0, sticky="we")
            ttk.Combobox(
                dlg, textvariable=trusteev, values=list(self.sidscombo.keys())
            ).grid(row=0, column=1, sticky="we")

            # ACE type
            ttk.Label(dlg, text="ACE type").grid(row=1, column=0, sticky="we")
            acetypefrm = ttk.Frame(
                dlg,
            )
            acetypev = tk.IntVar()
            acetypev.set(ace.AceType - 5 if ace.AceType >= 5 else ace.AceType)
            ttk.Radiobutton(
                acetypefrm,
                variable=acetypev,
                text="Allow",
                value=0x00,
            ).grid(row=0, column=0)
            ttk.Radiobutton(
                acetypefrm,
                variable=acetypev,
                text="Deny",
                value=0x01,
            ).grid(row=0, column=1)
            ttk.Radiobutton(
                acetypefrm,
                variable=acetypev,
                text="Audit",
                value=0x02,
            ).grid(row=0, column=2)
            ttk.Radiobutton(
                acetypefrm,
                variable=acetypev,
                text="Alarm",
                value=0x03,
                state=tk.DISABLED,
            ).grid(row=0, column=3)
            acetypefrm.grid(row=1, column=1, sticky="we")

            # Access Mask
            accessmaskfrm = ttk.LabelFrame(
                dlg,
                text="Access Mask",
            )
            sdvars = [None] * len(LDAP_DS_ACCESS_RIGHTS)
            for i, maskval in enumerate(LDAP_DS_ACCESS_RIGHTS.values()):
                sdvars[i] = tk.BooleanVar()
                sdvars[i].set(getattr(data["mask"], maskval))
                ttk.Checkbutton(accessmaskfrm, variable=sdvars[i], text=maskval).grid(
                    row=i // 4, column=i % 4, sticky="w"
                )
            accessmaskfrm.grid(row=2, column=0, columnspan=2, sticky="we")

            # ACE flags
            aceflagsfrm = ttk.LabelFrame(
                dlg,
                text="Access Mask",
            )
            aceflagsvars = [None] * len(WINNT_ACE_FLAGS)
            for i, aceval in enumerate(WINNT_ACE_FLAGS.values()):
                aceflagsvars[i] = tk.BooleanVar()
                aceflagsvars[i].set(getattr(ace.AceFlags, aceval))
                ttk.Checkbutton(
                    aceflagsfrm, variable=aceflagsvars[i], text=aceval
                ).grid(row=i // 4, column=i % 4, sticky="w")
            aceflagsfrm.grid(row=3, column=0, columnspan=2, sticky="we")

            # Object type
            objecttypev = tk.StringVar()
            objecttypev.set(self._rslvtype(data["object-guid"]) or "None")
            ttk.Label(dlg, text="Object type").grid(row=5, column=0, sticky="we")
            ttk.Combobox(
                dlg, textvariable=objecttypev, values=list(self.guidscombo.keys())
            ).grid(row=5, column=1, sticky="we")

            # Inherited object type
            inheritedobjecttypev = tk.StringVar()
            inheritedobjecttypev.set(
                self._rslvtype(data["inherited-object-guid"]) or "None"
            )
            ttk.Label(dlg, text="Inherited object type").grid(
                row=6, column=0, sticky="we"
            )
            ttk.Combobox(
                dlg,
                textvariable=inheritedobjecttypev,
                values=list(self.guidscomboobject.keys()),
            ).grid(row=6, column=1, sticky="we")

            # OK / Cancel
            btnfrm = ttk.Frame(dlg)
            ttk.Button(btnfrm, text="OK", command=subpopup.dismiss).grid(
                row=0, column=0
            )
            ttk.Button(btnfrm, text="Cancel", command=subpopup.cancel).grid(
                row=0, column=1
            )
            btnfrm.grid(row=7)

            # Setup
            if subpopup.run():
                # Cancelled
                return

            # Get values
            trustee = trusteev.get()
            acetype = acetypev.get()
            objecttype = objecttypev.get()
            inheritedobjecttype = inheritedobjecttypev.get()
            mask = 0
            for i, (sdvar, v) in enumerate(
                zip(sdvars, list(LDAP_DS_ACCESS_RIGHTS.keys()))
            ):
                if sdvar is None:
                    continue
                if sdvar.get():
                    mask |= v
            aceflags = 0
            for i, (aceflagvar, v) in enumerate(
                zip(aceflagsvars, list(WINNT_ACE_FLAGS.keys()))
            ):
                if aceflagvar is None:
                    continue
                if aceflagvar.get():
                    aceflags |= v

            # Set back into ACE
            if trustee in self.sidscombo:
                Sid = WINNT_SID.fromstr(self.sidscombo[trustee])
            else:
                Sid = WINNT_SID.fromstr(trustee)
            if objecttype in self.guidscombo:
                objecttype = self.guidscombo[objecttype]
            elif objecttype:
                objecttype = uuid.UUID(objecttype)
            if inheritedobjecttype in self.guidscomboobject:
                inheritedobjecttype = self.guidscomboobject[inheritedobjecttype]
            elif inheritedobjecttype:
                inheritedobjecttype = uuid.UUID(inheritedobjecttype)
            Flags = 0
            if objecttype:
                Flags |= 1
            if inheritedobjecttype:
                Flags |= 2
            if acetype == 0x00:
                if Flags:
                    ace.AceType = 0x05
                    ace.payload = WINNT_ACCESS_ALLOWED_OBJECT_ACE(
                        Mask=mask,
                        Sid=Sid,
                        Flags=Flags,
                        ObjectType=objecttype,
                        InheritedObjectType=inheritedobjecttype,
                    )
                else:
                    ace.AceType = 0x00
                    ace.payload = WINNT_ACCESS_ALLOWED_ACE(
                        Mask=mask,
                        Sid=Sid,
                    )
            elif acetype == 0x01:
                if Flags:
                    ace.AceType = 0x06
                    ace.payload = WINNT_ACCESS_DENIED_OBJECT_ACE(
                        Mask=mask,
                        Sid=Sid,
                        Flags=Flags,
                        ObjectType=objecttype,
                        InheritedObjectType=inheritedobjecttype,
                    )
                else:
                    ace.AceType = 0x01
                    ace.payload = WINNT_ACCESS_DENIED_ACE(
                        Mask=mask,
                        Sid=Sid,
                    )
            elif acetype == 0x02:
                if Flags:
                    ace.AceType = 0x07
                    ace.payload = WINNT_SYSTEM_AUDIT_OBJECT_ACE(
                        Mask=mask,
                        Sid=Sid,
                        Flags=Flags,
                        ObjectType=objecttype,
                        InheritedObjectType=inheritedobjecttype,
                    )
                else:
                    ace.AceType = 0x02
                    ace.payload = WINNT_SYSTEM_AUDIT_ACE(
                        Mask=mask,
                        Sid=Sid,
                    )
            else:
                raise NotImplementedError
            ace.AceFlags = aceflags

        def addace(id, table, ace, pos="end"):
            data = ace.extractData(accessMask=LDAP_DS_ACCESS_RIGHTS)
            table.insert(
                "",
                pos,
                id,
                values=(
                    ace.sprintf("%AceType%"),
                    self._rslvsid(data["sid-string"]),
                    str(data["mask"])
                    + (
                        " (%s)" % self._rslvtype(data["object-guid"])
                        if data["object-guid"]
                        else ""
                    ),
                    ace.sprintf("%AceFlags%"),
                ),
            )

        def acltable(name):
            aclfrm = ttk.LabelFrame(dlg, text=name, borderwidth=0)

            tvfr = ttk.Frame(aclfrm)
            tvfr.grid_columnconfigure(0, weight=1)
            tvfr.grid_rowconfigure(0, weight=1)

            acltree = ttk.Treeview(
                tvfr, show="headings", columns=("type", "trustee", "rights", "flags")
            )
            acltree.heading("type", text="Type")
            acltree.heading("trustee", text="Trustee")
            acltree.heading("rights", text="Rights")
            acltree.heading("flags", text="Flags")

            tree_scrollbar = AutoHideScrollbar(
                tvfr, orient="vertical", command=acltree.yview
            )
            acltree.configure(yscrollcommand=tree_scrollbar.set)
            acltree.grid(row=0, column=0, sticky="nsew")

            # Populate
            aclobj = getattr(nTSecurityDescriptor, name, None)
            if aclobj is not None:
                for i, ace in enumerate(aclobj.Aces):
                    addace(i, acltree, ace)

            def add(*_):
                ace = WINNT_ACE_HEADER() / WINNT_ACCESS_ALLOWED_ACE()
                acegui(ace)
                # Append
                aclobj.Aces.append(ace)
                addace(len(aclobj.Aces) - 1, acltree, ace)

            def delete(*_):
                try:
                    selected = int(acltree.selection()[0])
                    del aclobj.Aces[selected]
                except IndexError:
                    return
                # Full refresh as indexes change.
                acltree.delete(*acltree.get_children())
                for i, ace in enumerate(aclobj.Aces):
                    addace(i, acltree, ace)

            def edit(*_):
                try:
                    selected = int(acltree.selection()[0])
                    ace = aclobj.Aces[selected]
                except IndexError:
                    return
                acegui(ace)
                # Update
                acltree.delete(selected)
                addace(selected, acltree, ace, pos=selected)

            btnfrm = ttk.Frame(aclfrm)
            btnfrm.grid_columnconfigure(0, weight=1)
            ttk.Button(btnfrm, text="Add", command=add).grid(row=0)
            ttk.Button(btnfrm, text="Delete", command=delete).grid(row=1)
            ttk.Button(btnfrm, text="Edit", command=edit).grid(row=2)
            btnfrm.pack(side="right")

            tvfr.pack(fill="both", expand=True)
            return aclfrm

        acltable("DACL").grid(row=2, sticky="we")
        acltable("SACL").grid(row=3, sticky="we")

        btnfrm = ttk.Frame(dlg)
        ttk.Button(btnfrm, text="Update", command=popup.dismiss).grid(row=0, column=0)
        ttk.Button(btnfrm, text="Cancel", command=popup.cancel).grid(row=0, column=1)
        btnfrm.grid(row=4)

        # Setup
        if popup.run():
            # Cancelled
            return

        # From UI back into ntSecurityDescriptor

        # Owner
        owner = ownerv.get()
        if owner in self.sidscombo:
            nTSecurityDescriptor.OwnerSid = WINNT_SID.fromstr(self.sidscombo[owner])
        else:
            nTSecurityDescriptor.OwnerSid = WINNT_SID.fromstr(owner)

        # Group
        group = groupv.get()
        if group in self.sidscombo:
            nTSecurityDescriptor.GroupSid = WINNT_SID.fromstr(self.sidscombo[group])
        else:
            nTSecurityDescriptor.GroupSid = WINNT_SID.fromstr(group)

        # Control
        control = SECURITY_DESCRIPTOR(Control=0).Control
        for i, (sdvar, v) in enumerate(zip(sdvars, sdflags)):
            if sdvar is None:
                continue
            if sdvar.get():
                control |= v
        nTSecurityDescriptor.Control = control

        # Offsets need to be recalculated
        nTSecurityDescriptor.OwnerSidOffset = None
        nTSecurityDescriptor.GroupSidOffset = None
        nTSecurityDescriptor.DACLOffset = None
        nTSecurityDescriptor.SACLOffset = None

        # Pfew, we did it. That was some big UI.

        # Now update the SD.
        try:
            self.client.modify(
                object=item,
                changes=[
                    LDAP_ModifyRequestChange(
                        operation="replace",
                        modification=LDAP_PartialAttribute(
                            type="ntSecurityDescriptor",
                            values=[
                                LDAP_AttributeValue(value=bytes(nTSecurityDescriptor))
                            ],
                        ),
                    )
                ],
                controls=[
                    LDAP_Control(
                        controlType="1.2.840.113556.1.4.801",
                        criticality=True,
                        controlValue=LDAP_serverSDFlagsControl(
                            flags="OWNER+GROUP+DACL+SACL",
                        ),
                    )
                ],
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        self.tprint("Security descriptor updated.")

    def _members_popup(self, selection, mode="memberof"):
        """
        The base of the "Member Of" and "Members" popups

        :param mode: either "memberof" or "members"
        """
        # Get clicked item
        item = self.tk_tree.selection()[0]

        # Get the user attributes
        try:
            results = self.client.search(
                baseObject=item,
                scope=0,
                attributes=["objectClass", "memberOf"],
            )
            if item not in results:
                raise ValueError("Bad output")
            attributes = results[item]
        except ValueError as ex:
            self.tprint(str(ex))
            return
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        # Check that this item is indeed, a user or a group
        if not any(x in ["user", "group"] for x in attributes.get("objectClass", [])):
            messagebox.showerror("Error", "Object is neither a user nor a group !")
            return

        # Keep track of previous members, and changed ones
        og_members = set(attributes.get("memberOf", []))
        members = list(og_members)

        # Get dialog
        popup = BasePopup(self.root)
        dlg = popup.dlg

        # "Member Of" UI
        dlg.grid_rowconfigure(0, weight=1)
        dlg.grid_columnconfigure(0, weight=1)

        memberoffrm = ttk.LabelFrame(
            dlg,
            text="Member Of",
        )
        memberoffrm.grid_rowconfigure(0, weight=1)
        memberoffrm.grid_columnconfigure(0, weight=1)

        # Members list
        entrylist = tk.Listbox(memberoffrm)
        entrylist.grid(row=0, sticky="new")

        def add(*_, parentdlg=dlg):
            # Sub-dialog
            subpopup = BasePopup(parentdlg)
            dlg = subpopup.dlg

            # New group field
            newgroupv = tk.StringVar()
            ttk.Label(dlg, text="Group CN:").grid(row=0, sticky="we")
            newgroupf = tk.Entry(dlg, textvariable=newgroupv)
            newgroupf.grid(row=1, sticky="we")

            # OK / Cancel
            btnfrm = ttk.Frame(dlg)
            ttk.Button(btnfrm, text="OK", command=subpopup.dismiss).grid(
                row=0, column=0
            )
            ttk.Button(btnfrm, text="Cancel", command=subpopup.cancel).grid(
                row=0, column=1
            )
            btnfrm.grid(row=2, ipadx=5)

            # Focus
            newgroupf.focus()

            if subpopup.run():
                return

            # Get results
            newgroup = newgroupv.get()

            if newgroup:
                # Store
                members.append(newgroup)
                # Display
                entrylist.insert("end", newgroup)

        def delete(*_):
            try:
                selected = int(entrylist.curselection()[0])
            except IndexError:
                return
            # Drop
            del members[selected]
            # Remove from list
            entrylist.delete(selected)

        # Add / Delete
        btnfrm = ttk.Frame(memberoffrm)
        ttk.Button(btnfrm, text="Add", command=add).grid(row=0, column=0)
        ttk.Button(btnfrm, text="Delete", command=delete).grid(row=0, column=1)
        btnfrm.grid(row=1, sticky="we")

        # Populate
        for group in og_members:
            entrylist.insert("end", group)
            og_members.add(group)

        memberoffrm.grid(row=0, columnspan=2, sticky="we")

        # OK / Cancel
        btnfrm = ttk.Frame(dlg)
        ttk.Button(btnfrm, text="OK", command=popup.dismiss).grid(row=0, column=0)
        ttk.Button(btnfrm, text="Cancel", command=popup.cancel).grid(row=0, column=1)
        btnfrm.grid(row=1, ipadx=5)

        # Setup
        if popup.run():
            # Cancelled
            return

        # Get results
        members = set(members)
        to_add = members - og_members
        to_rem = og_members - members
        operations = [("add", x) for x in to_add] + [("delete", x) for x in to_rem]

        for op, group in operations:
            # Run the operations: on multiple groups, add/remove ourselves from "member"
            try:
                results = self.client.modify(
                    object=group,
                    changes=[
                        LDAP_ModifyRequestChange(
                            operation=op,
                            modification=LDAP_PartialAttribute(
                                type="member",
                                values=[LDAP_AttributeValue(value=item)],
                            ),
                        )
                    ],
                )
            except ValueError as ex:
                self.tprint(str(ex))
                return
            except LDAP_Exception as ex:
                self.tprint(
                    ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                    tags=["error"],
                )
                return

        self.tprint("Groups of '%s' updated !" % item)

    def editmemberof(self, *_):
        """
        Edit popup for "Member Of"
        """
        # Get clicked item
        item = self.tk_tree.selection()[0]

        self._members_popup(item, "memberof")

    def _edit_popup(self, selection, mode="edit", editattrs={}):
        """
        The base of the "Edit" and "Duplicate" popups

        :param mode: either "edit" or "new"
        :param editattrs: existing attributes to edit
        """
        # Get dialog
        popup = BasePopup(self.root)
        dlg = popup.dlg

        # Edit UI
        dlg.grid_columnconfigure(1, weight=1)

        # DN
        dnv = tk.StringVar()
        dnv.set(selection)
        if mode == "edit":
            ttk.Label(dlg, text="DN:").grid(row=0, column=0, sticky="w")
        else:
            ttk.Label(dlg, text="New DN:").grid(row=0, column=0, sticky="w")
        tk.Entry(dlg, textvariable=dnv).grid(row=0, column=1, sticky="we")

        # "Edit entry" sub-box
        editentryfrm = ttk.LabelFrame(
            dlg,
            text="Edit Entry",
        )
        attributev = tk.StringVar()
        ttk.Label(editentryfrm, text="Attribute:").grid(row=0, column=0)
        tk.Entry(editentryfrm, textvariable=attributev).grid(
            row=0, column=1, sticky="we"
        )

        valuesv = tk.StringVar()
        ttk.Label(editentryfrm, text="Values:").grid(row=1, column=0)
        tk.Entry(editentryfrm, textvariable=valuesv).grid(row=1, column=1, sticky="we")

        # "Operation" subbox: the radio + the buttons
        opsfrm = ttk.Frame(editentryfrm)
        operationfrm = ttk.LabelFrame(
            opsfrm,
            text="Operation",
        )
        scopev = tk.IntVar()
        scopev.set(0)
        ttk.Radiobutton(
            operationfrm,
            variable=scopev,
            text="Add",
            value=0,
        ).grid(row=0, column=0)
        ttk.Radiobutton(
            operationfrm,
            variable=scopev,
            text="Delete",
            value=1,
        ).grid(row=0, column=1)
        ttk.Radiobutton(
            operationfrm,
            variable=scopev,
            text="Replace",
            value=2,
        ).grid(row=0, column=2)
        operationfrm.grid(row=0, column=0, columnspan=2, sticky="we")

        if mode == "new":
            # In 'new', the only allowed operation is 'Add'
            for child in operationfrm.winfo_children():
                child.configure(state=tk.DISABLED)

        operations = []

        def enterentrylist():
            """
            This is called to add an element to the "Entry List"
            """
            op = scopev.get()
            attr = attributev.get()
            val = valuesv.get()
            ident = "[%s]%s:%s" % (
                {0: "Add", 1: "Delete", 2: "Replace"}[op],
                attr,
                val,
            )
            # Once we have an ident, actually parse the value entered by the user
            try:
                val = self._parse_attribute(attr, val)
            except ValueError:
                # Parsing failed, show a popup and return without clearing !
                return
            # Get current selection and reset it
            selected = self.currently_editing
            self.currently_editing = None
            # Do we have a selection
            if selected is not None:
                # Yes, edit
                # Set in storage
                operations[selected] = (op, attr, val)
                # Re-add to display
                entrylist.delete(selected)
                entrylist.insert(selected, ident)
                # Reset selection btw
                entrylist.itemconfigure(selected, fg="black")
                entrylist.see(selected)
            else:
                # No, create
                # Add to storage
                operations.append((op, attr, val))
                # Add to display
                entrylist.insert("end", ident)
            # Clear to really show we're done
            scopev.set(0)
            attributev.set("")
            valuesv.set("")

        def editentrylist():
            """
            This is called to load an element from the "Entry List"
            """
            try:
                selected = int(entrylist.curselection()[0])
            except IndexError:
                return
            # If there's a previously edited (unfinished), clear
            if self.currently_editing is not None:
                entrylist.itemconfigure(self.currently_editing, fg="black")
            # Set currently edited mode
            self.currently_editing = selected
            # Show selected item in blue
            entrylist.itemconfigure(selected, fg="blue")
            entrylist.selection_clear(selected)

            operation = operations[selected]
            # Set textboxes
            scopev.set(operation[0])
            attributev.set(operation[1])
            valuesv.set(self._format_attribute(operation[1], operation[2]))

        def removeentrylist():
            """
            This is called to remove an element from the "Entry List"
            """
            try:
                selected = entrylist.curselection()[0]
            except IndexError:
                return
            # Remove from storage
            del operations[selected]
            # Remove from display
            entrylist.delete(selected)

        ttk.Button(
            opsfrm,
            text="Enter",
            command=enterentrylist,
        ).grid(row=0, column=2)

        opsfrm.grid(row=2, column=0, columnspan=2)
        editentryfrm.grid(row=1, column=0, columnspan=2)

        # Entry list
        entrylistfrm = ttk.LabelFrame(
            dlg,
            text="Entry List",
        )
        entrylistfrm.grid_columnconfigure(0, weight=1)

        entrylist = tk.Listbox(entrylistfrm)
        entrylist.grid(row=0, sticky="we", padx=5)

        entrylistbtns = ttk.Frame(entrylistfrm)
        ttk.Button(
            entrylistbtns,
            text="Edit",
            command=editentrylist,
        ).pack(side="left")
        ttk.Button(
            entrylistbtns,
            text="Remove",
            command=removeentrylist,
        ).pack(side="right")
        entrylistbtns.grid(row=1, sticky="we", padx=10)

        entrylistfrm.grid(row=3, column=0, columnspan=2, sticky="we", pady=5)

        if mode == "new":
            for attr, val in editattrs.items():
                # Add to storage
                operations.append((0, attr, val))
                # Add to display
                ident = "[Add]%s:%s" % (
                    attr,
                    self._format_attribute(attr, val),
                )
                entrylist.insert("end", ident)

        # OK / Cancel
        btnfrm = ttk.Frame(dlg)
        ttk.Button(btnfrm, text="Run", command=popup.dismiss).pack(side="left")
        ttk.Button(btnfrm, text="Cancel", command=popup.cancel).pack(side="right")
        btnfrm.grid(row=4, column=0, columnspan=2, ipadx=10)

        # Setup
        if popup.run():
            # Cancelled
            return

        # Get values
        dn = dnv.get()

        return dn, operations

    def edit(self, *args):
        """
        Edit popup
        """
        # Get selected item
        try:
            selection = self.tk_tree.selection()[0]
        except IndexError:
            selection = ""

        results = self._edit_popup(selection)
        if not results:
            return
        dn, operations = results

        # Perform edit
        try:
            self.client.modify(
                object=dn,
                changes=[
                    LDAP_ModifyRequestChange(
                        operation=op,
                        modification=LDAP_PartialAttribute(
                            type=attr,
                            values=[LDAP_AttributeValue(value=x) for x in values],
                        ),
                    )
                    for (op, attr, values) in operations
                ],
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        self.tprint("Modify request succeeded.")

    def search(self, *args):
        """
        Search popup
        """
        # Get selected item
        try:
            selection = self.tk_tree.selection()[0]
        except IndexError:
            selection = "rootDSE"

        # Get dialog
        popup = BasePopup(self.root)
        dlg = popup.dlg

        # Search UI
        dlg.grid_columnconfigure(1, weight=1)

        basednv = tk.StringVar()
        basednv.set(selection)
        ttk.Label(dlg, text="Base DN").grid(row=0, column=0)
        basednf = tk.Entry(dlg, textvariable=basednv)
        basednf.grid(row=0, column=1, sticky="we")

        filterv = tk.StringVar()
        filterv.set(self.lastSearchString)
        ttk.Label(dlg, text="Filter").grid(row=1, column=0)
        tk.Entry(dlg, textvariable=filterv).grid(row=1, column=1, sticky="we")

        scopefrm = ttk.LabelFrame(
            dlg,
            text="Scope",
        )
        scopev = tk.IntVar()
        scopev.set(1)
        ttk.Radiobutton(
            scopefrm,
            variable=scopev,
            text="Base",
            value=0,
        ).grid(row=0, column=0)
        ttk.Radiobutton(
            scopefrm,
            variable=scopev,
            text="One Level",
            value=1,
        ).grid(row=0, column=1)
        ttk.Radiobutton(
            scopefrm,
            variable=scopev,
            text="Subtree",
            value=2,
        ).grid(row=0, column=2)
        scopefrm.grid(row=2, column=0, columnspan=2)

        ttk.Button(dlg, text="OK", command=popup.dismiss).grid(row=3, column=0)
        ttk.Button(dlg, text="Cancel", command=popup.cancel).grid(row=3, column=1)

        basednf.focus()

        # Setup
        if popup.run():
            # Cancelled
            return

        # Get values
        basedn = basednv.get()
        flt = filterv.get()
        scope = scopev.get()

        self.lastSearchString = flt

        # Perform search
        self.tprint("Searching...", flush=True)
        try:
            results = self.client.search(
                baseObject=basedn,
                scope=scope,
                filter=flt,
            )
        except ValueError as ex:
            self.tprint(str(ex))
            return
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        self.tprint("Getting %s entries..." % len(results))
        for item in results:
            self._showsearchresult(item, results)

    def modifydn(self, *args):
        """
        Modify the DN of an item
        """
        # Get selected item
        try:
            selection = self.tk_tree.selection()[0]
        except IndexError:
            selection = ""

        # Get dialog
        popup = BasePopup(self.root)
        dlg = popup.dlg

        # Duplicate UI
        dlg.grid_columnconfigure(1, weight=1)

        basednv = tk.StringVar()
        basednv.set(selection)
        ttk.Label(dlg, text="DN:").grid(row=0, column=0, sticky="w")
        basednf = tk.Entry(dlg, textvariable=basednv)
        basednf.grid(row=0, column=1, sticky="we")

        newdnv = tk.StringVar()
        ttk.Label(dlg, text="New DN:").grid(row=1, column=0, sticky="w")
        newdnf = tk.Entry(dlg, textvariable=newdnv)
        newdnf.grid(row=1, column=1, sticky="we")

        ttk.Button(dlg, text="OK", command=popup.dismiss).grid(row=2, column=0)
        ttk.Button(dlg, text="Cancel", command=popup.cancel).grid(row=2, column=1)

        if selection:
            newdnf.focus()
        else:
            basednf.focus()

        # Setup
        if popup.run():
            # Cancelled
            return

        # Get values
        basedn = basednv.get()
        newdn = newdnv.get()

        self.tprint("Changing %s to %s..." % (basedn, newdn))
        try:
            self.client.modifydn(
                entry=basedn,
                newdn=newdn,
            )
        except ValueError as ex:
            self.tprint(str(ex))
            return
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        self.tprint("OK !")

    def new(self, mode):
        """
        New popup. Called by both 'Add child' and 'Duplicate' popups
        """
        if mode == "duplicate":
            # Get selected item
            try:
                selection = self.tk_tree.selection()[0]
            except IndexError:
                selection = ""
        else:
            selection = ""

        existing_attributes = {}
        if selection:
            # Perform search to retrieve the attributes
            self.tprint("Getting attributes for %s..." % selection, flush=True)
            try:
                results = self.client.search(
                    baseObject=selection,
                    scope=0,
                )
                if selection not in results:
                    raise ValueError("Bad result")
                existing_attributes = results[selection]
            except ValueError as ex:
                self.tprint(str(ex))
                return
            except LDAP_Exception as ex:
                self.tprint(
                    ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                    tags=["error"],
                )
                return

        # Show edit popup to be able to change an attribute
        results = self._edit_popup(selection, mode="new", editattrs=existing_attributes)
        if not results:
            return
        newdn, changes = results

        # Extract all the 'add' attributes operations from changes
        attributes = {attr: val for (_, attr, val) in changes}

        self.tprint("Adding %s..." % newdn)
        try:
            self.client.add(
                newdn,
                attributes=attributes,
            )
        except LDAP_Exception as ex:
            self.tprint(
                ex.diagnosticMessage or "Error: %s" % ex.resultCode,
                tags=["error"],
            )
            return

        self.tprint("OK !")

    def duplicate(self, *args):
        return self.new("duplicate")

    def addchild(self, *args):
        return self.new("addchild")

    def _format_attribute(self, name, value, crop=False):
        """
        Format a LDAP attribute
        """
        if isinstance(value, list):
            # It's a list.
            return ";".join(self._format_attribute(name, v, crop=crop) for v in value)
        elif name == "objectSid":
            return WINNT_SID(value).summary()
        elif isinstance(value, bytes):
            # Catch-all for bytes values
            value = value.hex()
        else:
            # Catch-all
            value = str(value)
        # If cropping is enabled and requested, crop
        if crop and self.crop_output.get() and len(value) >= 80:
            return value[:80] + "... (%so)" % len(value)
        return value

    def _parse_attribute(self, name, value):
        """
        Parse a formatted attribute
        """
        parsed = []
        # Split across ;
        for val in value.split(";"):
            if name == "objectSid":
                val = WINNT_SID.fromstr(val)
            parsed.append(val)
        return parsed

    def tprint(self, x, tags=[], flush=False):
        """
        Print to text pane
        """
        self.tk_textpane.configure(state=tk.NORMAL)
        self.tk_textpane.insert("end", x + "\n", tuple(tags))
        self.tk_textpane.configure(state=tk.DISABLED)
        self.tk_textpane.see(tk.END)
        if flush:
            self.root.update()

    def main(self):
        """
        Main loop: start the GUI.
        """
        # Note: for TK doc, use https://tkdocs.com

        # Root
        self.root = tk.Tk()
        self.root.title("LDAPhero (@secdev/scapy)")
        self.root.option_add("*tearOff", False)

        # TTK style

        ttkstyle = ttk.Style()
        ttkstyle.theme_use("alt")
        ttkstyle.configure(
            "BorderFrame.TFrame",
            relief="groove",
            borderwidth=3,
        )

        # Global configuration variables
        self.crop_output = tk.BooleanVar()
        self.crop_output.set(True)

        # Create main frames, pack them in scrollable elements
        content = ttk.PanedWindow(self.root, orient=tk.HORIZONTAL)

        tvfr = ttk.Frame(content)
        tvfr.grid_columnconfigure(0, weight=1)
        tvfr.grid_rowconfigure(0, weight=1)
        self.tk_tree = ttk.Treeview(tvfr, show="tree")
        content.add(tvfr)
        self.tk_tree.bind("<Double-1>", self.treedoubleclick)

        tree_scrollbar = AutoHideScrollbar(
            tvfr, orient="vertical", command=self.tk_tree.yview
        )
        self.tk_tree.configure(yscrollcommand=tree_scrollbar.set)
        self.tk_tree.grid(row=0, column=0, sticky="nsew")
        self.tk_tree.column("#0", width=200)

        self.tk_textpane = tk.Text(content, state=tk.DISABLED)
        self.tk_textpane.tag_configure("bold", font="TkCaptionFont")
        self.tk_textpane.tag_configure("error", foreground="red")
        content.add(self.tk_textpane)

        # Menu
        menubar = tk.Menu(self.root)
        self.menu_connection = tk.Menu(menubar)
        self.menu_browse = tk.Menu(menubar)
        self.menu_view = tk.Menu(menubar)
        menubar.add_cascade(menu=self.menu_connection, label="Connection")
        self.menu_connection.add_command(label="Connect", command=self.connect)
        self.menu_connection.add_command(
            label="Bind", command=self.bind, state=tk.DISABLED, accelerator="Ctrl+B"
        )
        self.menu_connection.add_command(
            label="Disconnect", command=self.disconnect, state=tk.DISABLED
        )
        self.menu_connection.add_command(label="Quit", command=self.root.destroy)
        menubar.add_cascade(menu=self.menu_browse, label="Browse")
        self.menu_browse.add_command(
            label="Add child",
            command=self.addchild,
            state=tk.DISABLED,
            accelerator="Ctrl+A",
        )
        self.menu_browse.add_command(
            label="Modify", command=self.edit, state=tk.DISABLED, accelerator="Ctrl+M"
        )
        self.menu_browse.add_command(
            label="Modify DN",
            command=self.modifydn,
            state=tk.DISABLED,
            accelerator="Ctrl+R",
        )
        self.menu_browse.add_command(
            label="Search", command=self.search, state=tk.DISABLED, accelerator="Ctrl+S"
        )
        menubar.add_cascade(menu=self.menu_view, label="View")
        self.menu_view.add_command(
            label="Tree", command=self.tree, state=tk.DISABLED, accelerator="Ctrl+T"
        )
        self.menu_view.add_checkbutton(
            label="Crop output", onvalue=True, offvalue=False, variable=self.crop_output
        )
        self.root["menu"] = menubar

        # Right-click menu
        self.popup = tk.Menu(self.root, tearoff=0)
        self.popup.add_command(
            label="Search", command=self.search, accelerator="Ctrl+S"
        )
        self.popup.add_command(label="Modify", command=self.edit, accelerator="Ctrl+M")
        self.popup.add_command(
            label="Modify DN", command=self.modifydn, accelerator="Ctrl+R"
        )
        self.popup.add_command(label="Duplicate", command=self.duplicate)
        popup_adv = tk.Menu(self.popup)
        self.popup.add_cascade(label="Advanced", menu=popup_adv)
        popup_adv.add_command(label="Security descriptor", command=self.viewsec)
        popup_adv.add_command(label="Member Of", command=self.editmemberof)

        def do_popup(event):
            item = self.tk_tree.identify_row(event.y)
            if item:
                if self.tk_tree.tag_has("unclickable", item):
                    # Unclickable
                    return
                self.tk_tree.selection_set(item)
                self.popup.tk_popup(event.x_root, event.y_root)

        self.tk_tree.bind("<Button-3>", do_popup)

        # Shortcuts
        self.root.bind_all("<Control-b>", self.bind)
        self.root.bind_all("<Control-t>", self.tree)

        # Initial rendering
        content.pack(fill="both", expand=True)
        self.root.update()

        # Try connecting
        if self.host is not None:
            self.root.after(0, self.connect)

        # Main loop
        self.root.mainloop()



# SPDX-License-Identifier: BSD-3-Clause
# This file is part of Scapy
# See https://scapy.net/ for more information

"""
/*
 * Copyright (c) 1982, 1986, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)if_ether.h	8.1 (Berkeley) 6/10/93
 */
"""

# To quote Python's get-pip:

# Hi There!
#
# You may be wondering what this giant blob of binary data here is, you might
# even be worried that we're up to something nefarious (good for you for being
# paranoid!). This is a base85 encoding of a zip file, this zip file contains
# a version of '/etc/ethertypes', generated from OpenBSD's own copy, so that
# we are able to use it when not available on your OS.

# This file is automatically generated using
# scapy/tools/generate_ethertypes.py

import gzip
from base64 import b85decode

def _d(x: str) -> str:
    return gzip.decompress(
        b85decode(x.replace("\n", ""))
    ).decode()


DATA = _d("""
ABzY8N|hyM0{^91ZExbZ7XI#Eaioz}AWb2>6zJa3jGPeKXcNeglyY@-KbXWoE+IxqXv@Ff<IvJNyVvd
RMru`2KTnQ*-kxK=kS}1DTb^gUgmupL9Lm#y7x?k{3AafB>m=n6^CHTV6)&I=xJ;}8aq!6UL>!9?$pv
`GMJXbYp80SsD}m)4js=fFWN&Z9pC^&;iWd2T;Oc#8Qj`#hV;aMX!&)3O3HkNH4X`cC!>{f3)6-KcVH
s<QeA8w{k!-R(&&s0BU)Zm*<9@~S;x9lG&iU2I=)OS_{4K+y`7Yt#w*2}0pYQOEr3ouK-&?KL`On_<c
lct7y<|chvh?8HV;Dvs{@?Qj9NV@5F|8gPShT~#^zVITjnOp>4Z)J<;u$39a{5<La1I7F3`s`A&yz_S
8pk;=3J7zS6)7tdbliopL#BZWG6s6{rU60}8ziKohy1A#jh?r{kVsEWGInRT9ffhAGeRzh#*Z_utKvG
J!8!h;;zVPl-L&)O46+<vLk+6kd<)k*S$R&ZsTtJ4T@6?+zhG_4qcv<#gRPcdz}6j|1u9q<#np0&TXz
4j+iUG^alOkQc?vy3=YAbK&|qIf1b>P2SVhR?&?j@145>yMs!3G@=R9R6kif=#Vs(Z_r%4vh)K<>Hq+
<<{$+8p6phA&wP968%zdMFDXfV{V<mRtsU~DPggHE@n^MG7_1>P|&lZX{1Sy0z`Z)r!Lrsw6wsVMpW?
HK2ltJ-jLqx0sNmFysrtOQHs2a&&|tz=2rn|GRIdZ)S?i&94$))<;o{#?SHIG~#@{`Oxho^)8Z*Xts+
>08!2bkEWTZVwAL^809Umhnh-p#34`C5KFu7+M?bN<8PW<Q?Ctyu;7%$}`zuctImgcD$->&e(6(>3vI
0^nSOmOLV#1WJMBznTlw4ISAr-uKC_)eM`&ZWNVS{&wla*c6)G>vc$%3CL3_+lz20L{GM;1_te<703i
?`_lI^WSS$(VmP*k51VPUC0-X?v44uu1t2PkH(*J-3Atb2f5W<EZ!WPz7Kp>%>Y0p*g=mT&CA#?gLQG
nOiHyYraJt+m~t@zxV%GtwEUqJ4&4M%5Y*%gLH0kL?>Di_?FQ|Df#>3p6Bv4y1YER~}7d5RxDen3MKl
%l=PF){8<F!LndaJfe_2vz3m&oiG{AUNPBNw&cn$~l<@>TwVCUY`Z+8}O1S7f+~F(R&tk6?D(gdM{$>
Rn<4K#*sU<g`u+3ZU>iR>aI8mom)CpaNUWnS0o#jeZ};RS_A`+dJ44vVVpi<t6{4oP0AB^D2VY;QF7-
n#(IQ$|5&1452|KRNC4Z%1vK#EY77t=2Nd>_s9>i;mNIOV_R?23er;;32udbPPYetAO)8EQ`17Gf7XE
xTR#~jS#DYC0ZV@}Ed*NEwwe3d~neYn)M>#>D8)Mv#ZOs&hfi8v?oJXQkPgv{a;n8C$T7-sS&5nVt64
3CMka!ezgMt}S4aQ?-&d7Ld*IqO<wT(I}c2#QvJLS7F7g@3+o;`O$l|*;=9?z5VMKOw)VH3sxy3dkFI
VNI0ZDow2cBUb9e^Y30bhgn0_TP}DJK|GIDZnGc(&<5;MHt0_SeI2E2UA)*lCdrm9n%8|33&dPMA!($
BeGp4+_@b(ONDKsYh!a>CeMV{6fJ^!pV>J`AX&IdMSxL9KXbeelAWJWK~Z;XWKC==mrL15*J%=!MU$A
biCg2<ZRLTd={n`S{1;TnhM99#Zecb-t+3`zIPA&-{(~AmKf(1?$G}lBDq8;~cl6&%9*#QNLVCwec<N
OhmKqF8P+pvEf7)U&VIVDTU23X?x<0nJ1(&U)3KeMCOEtquZmEVvr>9HoDTYBRLpzO|C_&2yd7U9S8d
x8u6XzCe5C^HBn#8;J{Mv?fHQZ~T0kKTOV#{)L7MZw?8Zw=}F6I|`@;_cB9iCQFa!km^)NMjV)0p5O0
It9Wbs6j~N)eT^HLjgRUss%_=PMf&%F;P9u*ST~6hdA9j;chuibd1ImYp4qN$S<NtF8BPxssUBkQWnD
F&U$qCRv7lZBIj8;lIoec7~~f;fHt<+&bZd%3<qlx^>!Ng6a)JBa`D>F0hdWe=uaEi`5|7^7xoy<ESn
?*0PX={#UqjeBZeWLU{q9#KCAV9)Iov0;<qJeLo=cam}b<Pv>zQiLf)I5b|i7iBxP(mEZtL^N^HVfqK
C4iRV~;jg|flR!@$t_-A~S5(GomD)aR0p%!kR2I@NomVW!P0cT<_uPI-J%$u+d+}VRdhdoI{H!^yy9*
dz!#na_nk<X=;-}YF(<9U<#e(i8ahOW)Y<qfft(+$o!tCRd-W@8cYU3I6UzOFu1%Q63}nV&A58unUyf
%2RbGM)EFiI+9304%UFfb)cx46Ei!h!Wkw4sF>$k`1Y>R9-RYf3VA$lCJ?Ts%S*%w&BF4{>)YAMz+=w
*xr_asB;yN6Dpn6q|b<UwHk}oTKi9D`m;tb`p5<hhN{OFC5v{lZo96UTceSs3s5)7cG#Uxypr6JsG8?
xi1$siwk3P`kwrfx{4t4<(L|e8QlwYYk+g0a+^K{&{mCav{ficYJI*Ap%%4%uFtE%{gW&zhaigCotIL
qaMC5x^*QeK^O^v+f*XnNMHCN9=Np+<jT7Lay-GBZvWNl-a_>=dIZJd|iQBr*L^VHN1f55%`jsx>}L6
<P-b`7X_NHCi!!f`?;IvE<E5`PAoA2^CPK^9J^45O2YRs)*m6e`G-%d(f9M+|!lJh2@=9txZ;RQWT(M
>}0SC;Q_b9$A{i@cIQy_4UqIdGU$?!ejC~XT@GkQ5paM
""")



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2019 Luis Rosa <lmrosa@dei.uc.pt>

# scapy.contrib.description = PCOM Protocol
# scapy.contrib.status = loads

"""
PCOM

PCOM is a protocol to communicate with Unitronics PLCs either by serial
or TCP. Two modes are available, ASCII and Binary.

https://unitronicsplc.com/Download/SoftwareUtilities/Unitronics%20PCOM%20Protocol.pdf
"""

import struct

from scapy.packet import Packet, bind_layers
from scapy.layers.inet import TCP
from scapy.fields import XShortField, ByteEnumField, XByteField, \
    StrFixedLenField, StrLenField, LEShortField, \
    LEFieldLenField, XLE3BytesField, XLEShortField
from scapy.volatile import RandShort
from scapy.compat import bytes_encode, orb

_protocol_modes = {0x65: "ascii", 0x66: "binary"}

_ascii_command_codes = {
    "ID": "Send Identification Command",
    "CCR": "Send Start Command",
    "CCS": "Send Stop Command",
    "CCE": "Send Reset Command",
    "CCI": "Send Init Command",
    "CC": "Reply of Admin Commands (CC*)",
    "UG": "Get UnitID",
    "US": "Set UnitID",
    "RC": "Get RTC",
    "SC": "Set RTC",
    "RE": "Read Inputs",
    "RA": "Read Outputs",
    "GS": "Read System Bits",
    "GF": "Read System Integers",
    "RNH": "Read System Longs",
    "RNJ": "Read System Double Words",
    "RB": "Read Memory Bits",
    "RW": "Read Memory Integers",
    "RNL": "Read Memory Longs",
    "RND": "Read Memory Double Words",
    "RN": "Read Longs / Double Words",
    "SA": "Write Outputs",
    "SS": "Write System Bits",
    "SF": "Write System Integers",
    "SNH": "Write System Longs",
    "SNJ": "Write System Double Words",
    "SB": "Write Memory Bits",
    "SW": "Write Memory Integers",
    "SNL": "Write Memory Longs",
    "SND": "Write Memory Double Words",
    "SN": "Write Longs / Double Words"
}

_binary_command_codes = {
    0x0c: "Get PLC Name Request",
    0x8c: "Get PLC Name Reply",
    0x4d: "Read Operands Request",
    0xcd: "Read Operands Reply",
    0x04: "Read Data Table Request",
    0x84: "Read Data Table Reply",
    0x44: "Write Data Table Request",
    0xc4: "Write Data Table Reply"
}


class PCOM(Packet):
    fields_desc = [
        XShortField("transId", RandShort()),
        ByteEnumField("mode", 0x65, _protocol_modes),
        XByteField("reserved", 0x00),
        LEShortField("len", None)
    ]

    def post_build(self, pkt, pay):
        if self.len is None and pay:
            pkt = pkt[:4] + struct.pack("<H", len(pay))
        return pkt + pay


class PCOMRequest(PCOM):
    name = "PCOM/TCP Request"


class PCOMResponse(PCOM):
    name = "PCOM/TCP Response"


class PCOMAscii(Packet):
    @staticmethod
    def pcom_ascii_checksum(command):
        n = 0
        command = bytes_encode(command)
        for _, c in enumerate(command):
            n += orb(c)
        return list(map(ord, hex(n % 256)[2:].zfill(2).upper()))


class PCOMAsciiCommandField(StrLenField):
    def i2repr(self, pkt, x):
        s = super(PCOMAsciiCommandField, self).i2repr(pkt, x)
        code = s[1:4]  # check for 3 chars known codes
        if code in _ascii_command_codes:
            return _ascii_command_codes[code] + " " + s
        code = s[1:3]  # check for 2 chars known codes
        if code in _ascii_command_codes:
            return _ascii_command_codes[code] + " " + s
        return s


class PCOMAsciiRequest(PCOMAscii):
    name = "PCOM/ASCII Request"
    fields_desc = [
        StrFixedLenField("stx", "/", 1),
        StrFixedLenField("unitId", "00", 2),
        PCOMAsciiCommandField(
            "command", '', length_from=lambda pkt: pkt.underlayer.len - 6),
        XShortField("chksum", None),
        XByteField("etx", 0x0d)
    ]

    def post_build(self, pkt, pay):
        if self.chksum is None:
            chksum = PCOMAscii.pcom_ascii_checksum(pkt[1:-3])
            pkt = pkt[:-3] + struct.pack("2B", chksum[0], chksum[1]) + pkt[-1:]
        return pkt + pay


class PCOMAsciiResponse(PCOMAscii):
    name = "PCOM/ASCII Response"
    fields_desc = [
        StrFixedLenField("stx", "/A", 2),
        StrFixedLenField("unitId", "00", 2),
        PCOMAsciiCommandField(
            "command", '', length_from=lambda pkt: pkt.underlayer.len - 7),
        XShortField("chksum", None),
        XByteField("etx", 0x0d)
    ]

    def post_build(self, pkt, pay):
        if self.chksum is None:
            chksum = PCOMAscii.pcom_ascii_checksum(pkt[2:-3])
            pkt = pkt[:-3] + struct.pack("2B", chksum[0], chksum[1]) + pkt[-1:]
        return pkt + pay


class PCOMBinary(Packet):
    @staticmethod
    def pcom_binary_checksum(command):
        n = 0
        command = bytes_encode(command)
        for _, c in enumerate(command):
            c = c if isinstance(c, int) else ord(c)  # python 2 fallback
            n += c
        if n == 0:
            return [0x00, 0x00]
        else:
            two_complement = hex(0x10000 - (n % 0x10000))[2:].zfill(4)
            return [int(two_complement[:2], 16), int(two_complement[2:], 16)]

    def post_build(self, pkt, pay):
        if self.headerChksum is None:
            chksum = PCOMBinaryRequest.pcom_binary_checksum(pkt[:21])
            pkt = pkt[:22] + struct.pack("2B", chksum[1], chksum[0]) + pkt[24:]
        if self.footerChksum is None:
            chksum = PCOMBinaryRequest.pcom_binary_checksum(pkt[24:-3])
            pkt = pkt[:-3] + struct.pack("2B", chksum[1], chksum[0]) + pkt[-1:]
        return pkt + pay


class PCOMBinaryCommandField(XByteField):
    def i2repr(self, pkt, x):
        s = super(PCOMBinaryCommandField, self).i2repr(pkt, x)
        if x in _binary_command_codes:
            return _binary_command_codes[x] + " - " + s
        else:
            return s


class PCOMBinaryRequest(PCOMBinary):
    name = "PCOM/Binary Request"
    fields_desc = [
        StrFixedLenField("stx", "/_OPLC", 6),
        XByteField("id", 0x0),
        XByteField("reserved1", 0xfe),
        XByteField("reserved2", 0x1),
        XLE3BytesField("reserved3", 0x0),
        PCOMBinaryCommandField("command", None),
        XByteField("reserved4", 0x0),
        StrFixedLenField("commandSpecific", '', 6),
        LEFieldLenField("len", 0, length_of="data"),
        XLEShortField("headerChksum", None),
        StrLenField("data", '', length_from=lambda pkt: pkt.len),
        XLEShortField("footerChksum", None),
        XByteField("etx", 0x5c)
    ]


class PCOMBinaryResponse(PCOMBinary):
    name = "PCOM/Binary Response"
    fields_desc = [
        StrFixedLenField("stx", "/_OPLC", 6),
        XByteField("reserved1", 0xfe),
        XByteField("id", 0x0),
        XByteField("reserved2", 0x1),
        XLE3BytesField("reserved3", 0x0),
        PCOMBinaryCommandField("command", None),
        XByteField("reserved4", 0x0),
        StrFixedLenField("commandSpecific", '', 6),
        LEFieldLenField("len", 0, length_of="data"),
        XLEShortField("headerChksum", None),
        StrLenField("data", '', length_from=lambda pkt: pkt.len),
        XLEShortField("footerChksum", None),
        XByteField("etx", 0x5c)
    ]


bind_layers(TCP, PCOMRequest, dport=20256)
bind_layers(TCP, PCOMResponse, sport=20256)
bind_layers(PCOMRequest, PCOMAsciiRequest, mode=0x65)
bind_layers(PCOMRequest, PCOMBinaryRequest, mode=0x66)
bind_layers(PCOMResponse, PCOMAsciiResponse, mode=0x65)
bind_layers(PCOMResponse, PCOMBinaryResponse, mode=0x66)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# Acknowledgment: Maxence Tury <maxence.tury@ssi.gouv.fr>

"""
Classes that implement ASN.1 data structures.
"""

import copy

from functools import reduce

from scapy.asn1.asn1 import (
    ASN1_BIT_STRING,
    ASN1_BOOLEAN,
    ASN1_Class,
    ASN1_Class_UNIVERSAL,
    ASN1_Error,
    ASN1_INTEGER,
    ASN1_NULL,
    ASN1_OID,
    ASN1_Object,
    ASN1_STRING,
)
from scapy.asn1.ber import (
    BER_Decoding_Error,
    BER_id_dec,
    BER_tagging_dec,
    BER_tagging_enc,
)
from scapy.base_classes import BasePacket
from scapy.volatile import (
    GeneralizedTime,
    RandChoice,
    RandInt,
    RandNum,
    RandOID,
    RandString,
    RandField,
)

from scapy import packet

from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from scapy.asn1packet import ASN1_Packet


class ASN1F_badsequence(Exception):
    pass


class ASN1F_element(object):
    pass


##########################
#    Basic ASN1 Field    #
##########################

_I = TypeVar('_I')  # Internal storage
_A = TypeVar('_A')  # ASN.1 object


class ASN1F_field(ASN1F_element, Generic[_I, _A]):
    holds_packets = 0
    islist = 0
    ASN1_tag = ASN1_Class_UNIVERSAL.ANY
    context = ASN1_Class_UNIVERSAL  # type: Type[ASN1_Class]

    def __init__(self,
                 name,  # type: str
                 default,  # type: Optional[_A]
                 context=None,  # type: Optional[Type[ASN1_Class]]
                 implicit_tag=None,  # type: Optional[int]
                 explicit_tag=None,  # type: Optional[int]
                 flexible_tag=False,  # type: Optional[bool]
                 size_len=None,  # type: Optional[int]
                 ):
        # type: (...) -> None
        if context is not None:
            self.context = context
        self.name = name
        if default is None:
            self.default = default  # type: Optional[_A]
        elif isinstance(default, ASN1_NULL):
            self.default = default  # type: ignore
        else:
            self.default = self.ASN1_tag.asn1_object(default)  # type: ignore
        self.size_len = size_len
        self.flexible_tag = flexible_tag
        if (implicit_tag is not None) and (explicit_tag is not None):
            err_msg = "field cannot be both implicitly and explicitly tagged"
            raise ASN1_Error(err_msg)
        self.implicit_tag = implicit_tag and int(implicit_tag)
        self.explicit_tag = explicit_tag and int(explicit_tag)
        # network_tag gets useful for ASN1F_CHOICE
        self.network_tag = int(implicit_tag or explicit_tag or self.ASN1_tag)
        self.owners = []  # type: List[Type[ASN1_Packet]]

    def register_owner(self, cls):
        # type: (Type[ASN1_Packet]) -> None
        self.owners.append(cls)

    def i2repr(self, pkt, x):
        # type: (ASN1_Packet, _I) -> str
        return repr(x)

    def i2h(self, pkt, x):
        # type: (ASN1_Packet, _I) -> Any
        return x

    def m2i(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> Tuple[_A, bytes]
        """
        The good thing about safedec is that it may still decode ASN1
        even if there is a mismatch between the expected tag (self.ASN1_tag)
        and the actual tag; the decoded ASN1 object will simply be put
        into an ASN1_BADTAG object. However, safedec prevents the raising of
        exceptions needed for ASN1F_optional processing.
        Thus we use 'flexible_tag', which should be False with ASN1F_optional.

        Regarding other fields, we might need to know whether encoding went
        as expected or not. Noticeably, input methods from cert.py expect
        certain exceptions to be raised. Hence default flexible_tag is False.
        """
        diff_tag, s = BER_tagging_dec(s, hidden_tag=self.ASN1_tag,
                                      implicit_tag=self.implicit_tag,
                                      explicit_tag=self.explicit_tag,
                                      safe=self.flexible_tag,
                                      _fname=self.name)
        if diff_tag is not None:
            # this implies that flexible_tag was True
            if self.implicit_tag is not None:
                self.implicit_tag = diff_tag
            elif self.explicit_tag is not None:
                self.explicit_tag = diff_tag
        codec = self.ASN1_tag.get_codec(pkt.ASN1_codec)
        if self.flexible_tag:
            return codec.safedec(s, context=self.context)  # type: ignore
        else:
            return codec.dec(s, context=self.context)  # type: ignore

    def i2m(self, pkt, x):
        # type: (ASN1_Packet, Union[bytes, _I, _A]) -> bytes
        if x is None:
            return b""
        if isinstance(x, ASN1_Object):
            if (self.ASN1_tag == ASN1_Class_UNIVERSAL.ANY or
                x.tag == ASN1_Class_UNIVERSAL.RAW or
                x.tag == ASN1_Class_UNIVERSAL.ERROR or
               self.ASN1_tag == x.tag):
                s = x.enc(pkt.ASN1_codec)
            else:
                raise ASN1_Error("Encoding Error: got %r instead of an %r for field [%s]" % (x, self.ASN1_tag, self.name))  # noqa: E501
        else:
            s = self.ASN1_tag.get_codec(pkt.ASN1_codec).enc(x, size_len=self.size_len)
        return BER_tagging_enc(s,
                               implicit_tag=self.implicit_tag,
                               explicit_tag=self.explicit_tag)

    def any2i(self, pkt, x):
        # type: (ASN1_Packet, Any) -> _I
        return cast(_I, x)

    def extract_packet(self,
                       cls,  # type: Type[ASN1_Packet]
                       s,  # type: bytes
                       _underlayer=None  # type: Optional[ASN1_Packet]
                       ):
        # type: (...) -> Tuple[ASN1_Packet, bytes]
        try:
            c = cls(s, _underlayer=_underlayer)
        except ASN1F_badsequence:
            c = packet.Raw(s, _underlayer=_underlayer)  # type: ignore
        cpad = c.getlayer(packet.Raw)
        s = b""
        if cpad is not None:
            s = cpad.load
            if cpad.underlayer:
                del cpad.underlayer.payload
        return c, s

    def build(self, pkt):
        # type: (ASN1_Packet) -> bytes
        return self.i2m(pkt, getattr(pkt, self.name))

    def dissect(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> bytes
        v, s = self.m2i(pkt, s)
        self.set_val(pkt, v)
        return s

    def do_copy(self, x):
        # type: (Any) -> Any
        if isinstance(x, list):
            x = x[:]
            for i in range(len(x)):
                if isinstance(x[i], BasePacket):
                    x[i] = x[i].copy()
            return x
        if hasattr(x, "copy"):
            return x.copy()
        return x

    def set_val(self, pkt, val):
        # type: (ASN1_Packet, Any) -> None
        setattr(pkt, self.name, val)

    def is_empty(self, pkt):
        # type: (ASN1_Packet) -> bool
        return getattr(pkt, self.name) is None

    def get_fields_list(self):
        # type: () -> List[ASN1F_field[Any, Any]]
        return [self]

    def __str__(self):
        # type: () -> str
        return repr(self)

    def randval(self):
        # type: () -> RandField[_I]
        return cast(RandField[_I], RandInt())

    def copy(self):
        # type: () -> ASN1F_field[_I, _A]
        return copy.copy(self)


############################
#    Simple ASN1 Fields    #
############################

class ASN1F_BOOLEAN(ASN1F_field[bool, ASN1_BOOLEAN]):
    ASN1_tag = ASN1_Class_UNIVERSAL.BOOLEAN

    def randval(self):
        # type: () -> RandChoice
        return RandChoice(True, False)


class ASN1F_INTEGER(ASN1F_field[int, ASN1_INTEGER]):
    ASN1_tag = ASN1_Class_UNIVERSAL.INTEGER

    def randval(self):
        # type: () -> RandNum
        return RandNum(-2**64, 2**64 - 1)


class ASN1F_enum_INTEGER(ASN1F_INTEGER):
    def __init__(self,
                 name,  # type: str
                 default,  # type: ASN1_INTEGER
                 enum,  # type: Dict[int, str]
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[Any]
                 explicit_tag=None,  # type: Optional[Any]
                 ):
        # type: (...) -> None
        super(ASN1F_enum_INTEGER, self).__init__(
            name, default, context=context,
            implicit_tag=implicit_tag,
            explicit_tag=explicit_tag
        )
        i2s = self.i2s = {}  # type: Dict[int, str]
        s2i = self.s2i = {}  # type: Dict[str, int]
        if isinstance(enum, list):
            keys = range(len(enum))
        else:
            keys = list(enum)
        if any(isinstance(x, str) for x in keys):
            i2s, s2i = s2i, i2s  # type: ignore
        for k in keys:
            i2s[k] = enum[k]
            s2i[enum[k]] = k

    def i2m(self,
            pkt,  # type: ASN1_Packet
            s,  # type: Union[bytes, str, int, ASN1_INTEGER]
            ):
        # type: (...) -> bytes
        if not isinstance(s, str):
            vs = s
        else:
            vs = self.s2i[s]
        return super(ASN1F_enum_INTEGER, self).i2m(pkt, vs)

    def i2repr(self,
               pkt,  # type: ASN1_Packet
               x,  # type: Union[str, int]
               ):
        # type: (...) -> str
        if x is not None and isinstance(x, ASN1_INTEGER):
            r = self.i2s.get(x.val)
            if r:
                return "'%s' %s" % (r, repr(x))
        return repr(x)


class ASN1F_BIT_STRING(ASN1F_field[str, ASN1_BIT_STRING]):
    ASN1_tag = ASN1_Class_UNIVERSAL.BIT_STRING

    def __init__(self,
                 name,  # type: str
                 default,  # type: Optional[Union[ASN1_BIT_STRING, AnyStr]]
                 default_readable=True,  # type: bool
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[int]
                 explicit_tag=None,  # type: Optional[int]
                 ):
        # type: (...) -> None
        super(ASN1F_BIT_STRING, self).__init__(
            name, None, context=context,
            implicit_tag=implicit_tag,
            explicit_tag=explicit_tag
        )
        if isinstance(default, (bytes, str)):
            self.default = ASN1_BIT_STRING(default,
                                           readable=default_readable)
        else:
            self.default = default

    def randval(self):
        # type: () -> RandString
        return RandString(RandNum(0, 1000))


class ASN1F_STRING(ASN1F_field[str, ASN1_STRING]):
    ASN1_tag = ASN1_Class_UNIVERSAL.STRING

    def randval(self):
        # type: () -> RandString
        return RandString(RandNum(0, 1000))


class ASN1F_NULL(ASN1F_INTEGER):
    ASN1_tag = ASN1_Class_UNIVERSAL.NULL


class ASN1F_OID(ASN1F_field[str, ASN1_OID]):
    ASN1_tag = ASN1_Class_UNIVERSAL.OID

    def randval(self):
        # type: () -> RandOID
        return RandOID()


class ASN1F_ENUMERATED(ASN1F_enum_INTEGER):
    ASN1_tag = ASN1_Class_UNIVERSAL.ENUMERATED


class ASN1F_UTF8_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.UTF8_STRING


class ASN1F_NUMERIC_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.NUMERIC_STRING


class ASN1F_PRINTABLE_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.PRINTABLE_STRING


class ASN1F_T61_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.T61_STRING


class ASN1F_VIDEOTEX_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.VIDEOTEX_STRING


class ASN1F_IA5_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.IA5_STRING


class ASN1F_GENERAL_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.GENERAL_STRING


class ASN1F_UTC_TIME(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.UTC_TIME

    def randval(self):  # type: ignore
        # type: () -> GeneralizedTime
        return GeneralizedTime()


class ASN1F_GENERALIZED_TIME(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.GENERALIZED_TIME

    def randval(self):  # type: ignore
        # type: () -> GeneralizedTime
        return GeneralizedTime()


class ASN1F_ISO646_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.ISO646_STRING


class ASN1F_UNIVERSAL_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.UNIVERSAL_STRING


class ASN1F_BMP_STRING(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.BMP_STRING


class ASN1F_SEQUENCE(ASN1F_field[List[Any], List[Any]]):
    # Here is how you could decode a SEQUENCE
    # with an unknown, private high-tag prefix :
    # class PrivSeq(ASN1_Packet):
    #     ASN1_codec = ASN1_Codecs.BER
    #     ASN1_root = ASN1F_SEQUENCE(
    #                       <asn1 field #0>,
    #                       ...
    #                       <asn1 field #N>,
    #                       explicit_tag=0,
    #                       flexible_tag=True)
    # Because we use flexible_tag, the value of the explicit_tag does not matter.  # noqa: E501
    ASN1_tag = ASN1_Class_UNIVERSAL.SEQUENCE
    holds_packets = 1

    def __init__(self, *seq, **kwargs):
        # type: (*Any, **Any) -> None
        name = "dummy_seq_name"
        default = [field.default for field in seq]
        super(ASN1F_SEQUENCE, self).__init__(
            name, default, **kwargs
        )
        self.seq = seq
        self.islist = len(seq) > 1

    def __repr__(self):
        # type: () -> str
        return "<%s%r>" % (self.__class__.__name__, self.seq)

    def is_empty(self, pkt):
        # type: (ASN1_Packet) -> bool
        return all(f.is_empty(pkt) for f in self.seq)

    def get_fields_list(self):
        # type: () -> List[ASN1F_field[Any, Any]]
        return reduce(lambda x, y: x + y.get_fields_list(),
                      self.seq, [])

    def m2i(self, pkt, s):
        # type: (Any, bytes) -> Tuple[Any, bytes]
        """
        ASN1F_SEQUENCE behaves transparently, with nested ASN1_objects being
        dissected one by one. Because we use obj.dissect (see loop below)
        instead of obj.m2i (as we trust dissect to do the appropriate set_vals)
        we do not directly retrieve the list of nested objects.
        Thus m2i returns an empty list (along with the proper remainder).
        It is discarded by dissect() and should not be missed elsewhere.
        """
        diff_tag, s = BER_tagging_dec(s, hidden_tag=self.ASN1_tag,
                                      implicit_tag=self.implicit_tag,
                                      explicit_tag=self.explicit_tag,
                                      safe=self.flexible_tag,
                                      _fname=pkt.name)
        if diff_tag is not None:
            if self.implicit_tag is not None:
                self.implicit_tag = diff_tag
            elif self.explicit_tag is not None:
                self.explicit_tag = diff_tag
        codec = self.ASN1_tag.get_codec(pkt.ASN1_codec)
        i, s, remain = codec.check_type_check_len(s)
        if len(s) == 0:
            for obj in self.seq:
                obj.set_val(pkt, None)
        else:
            for obj in self.seq:
                try:
                    s = obj.dissect(pkt, s)
                except ASN1F_badsequence:
                    break
            if len(s) > 0:
                raise BER_Decoding_Error("unexpected remainder", remaining=s)
        return [], remain

    def dissect(self, pkt, s):
        # type: (Any, bytes) -> bytes
        _, x = self.m2i(pkt, s)
        return x

    def build(self, pkt):
        # type: (ASN1_Packet) -> bytes
        s = reduce(lambda x, y: x + y.build(pkt),
                   self.seq, b"")
        return super(ASN1F_SEQUENCE, self).i2m(pkt, s)


class ASN1F_SET(ASN1F_SEQUENCE):
    ASN1_tag = ASN1_Class_UNIVERSAL.SET


_SEQ_T = Union[
    'ASN1_Packet',
    Type[ASN1F_field[Any, Any]],
    'ASN1F_PACKET',
    ASN1F_field[Any, Any],
]


class ASN1F_SEQUENCE_OF(ASN1F_field[List[_SEQ_T],
                                    List[ASN1_Object[Any]]]):
    """
    Two types are allowed as cls: ASN1_Packet, ASN1F_field
    """
    ASN1_tag = ASN1_Class_UNIVERSAL.SEQUENCE
    islist = 1

    def __init__(self,
                 name,  # type: str
                 default,  # type: Any
                 cls,  # type: _SEQ_T
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[Any]
                 explicit_tag=None,  # type: Optional[Any]
                 ):
        # type: (...) -> None
        if isinstance(cls, type) and issubclass(cls, ASN1F_field) or \
                isinstance(cls, ASN1F_field):
            if isinstance(cls, type):
                self.fld = cls(name, b"")
            else:
                self.fld = cls
            self._extract_packet = lambda s, pkt: self.fld.m2i(pkt, s)
            self.holds_packets = 0
        elif hasattr(cls, "ASN1_root") or callable(cls):
            self.cls = cast("Type[ASN1_Packet]", cls)
            self._extract_packet = lambda s, pkt: self.extract_packet(
                self.cls, s, _underlayer=pkt)
            self.holds_packets = 1
        else:
            raise ValueError("cls should be an ASN1_Packet or ASN1_field")
        super(ASN1F_SEQUENCE_OF, self).__init__(
            name, None, context=context,
            implicit_tag=implicit_tag, explicit_tag=explicit_tag
        )
        self.default = default

    def is_empty(self,
                 pkt,  # type: ASN1_Packet
                 ):
        # type: (...) -> bool
        return ASN1F_field.is_empty(self, pkt)

    def m2i(self,
            pkt,  # type: ASN1_Packet
            s,  # type: bytes
            ):
        # type: (...) -> Tuple[List[Any], bytes]
        diff_tag, s = BER_tagging_dec(s, hidden_tag=self.ASN1_tag,
                                      implicit_tag=self.implicit_tag,
                                      explicit_tag=self.explicit_tag,
                                      safe=self.flexible_tag)
        if diff_tag is not None:
            if self.implicit_tag is not None:
                self.implicit_tag = diff_tag
            elif self.explicit_tag is not None:
                self.explicit_tag = diff_tag
        codec = self.ASN1_tag.get_codec(pkt.ASN1_codec)
        i, s, remain = codec.check_type_check_len(s)
        lst = []
        while s:
            c, s = self._extract_packet(s, pkt)  # type: ignore
            if c:
                lst.append(c)
        if len(s) > 0:
            raise BER_Decoding_Error("unexpected remainder", remaining=s)
        return lst, remain

    def build(self, pkt):
        # type: (ASN1_Packet) -> bytes
        val = getattr(pkt, self.name)
        if isinstance(val, ASN1_Object) and \
                val.tag == ASN1_Class_UNIVERSAL.RAW:
            s = cast(Union[List[_SEQ_T], bytes], val)
        elif val is None:
            s = b""
        else:
            s = b"".join(bytes(i) for i in val)
        return self.i2m(pkt, s)

    def i2repr(self, pkt, x):
        # type: (ASN1_Packet, _I) -> str
        if self.holds_packets:
            return super(ASN1F_SEQUENCE_OF, self).i2repr(pkt, x)  # type: ignore
        elif x is None:
            return "[]"
        else:
            return "[%s]" % ", ".join(
                self.fld.i2repr(pkt, x) for x in x  # type: ignore
            )

    def randval(self):
        # type: () -> Any
        if self.holds_packets:
            return packet.fuzz(self.cls())
        else:
            return self.fld.randval()

    def __repr__(self):
        # type: () -> str
        return "<%s %s>" % (self.__class__.__name__, self.name)


class ASN1F_SET_OF(ASN1F_SEQUENCE_OF):
    ASN1_tag = ASN1_Class_UNIVERSAL.SET


class ASN1F_IPADDRESS(ASN1F_STRING):
    ASN1_tag = ASN1_Class_UNIVERSAL.IPADDRESS


class ASN1F_TIME_TICKS(ASN1F_INTEGER):
    ASN1_tag = ASN1_Class_UNIVERSAL.TIME_TICKS


#############################
#    Complex ASN1 Fields    #
#############################

class ASN1F_optional(ASN1F_element):
    """
    ASN.1 field that is optional.
    """
    def __init__(self, field):
        # type: (ASN1F_field[Any, Any]) -> None
        field.flexible_tag = False
        self._field = field

    def __getattr__(self, attr):
        # type: (str) -> Optional[Any]
        return getattr(self._field, attr)

    def m2i(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> Tuple[Any, bytes]
        try:
            return self._field.m2i(pkt, s)
        except (ASN1_Error, ASN1F_badsequence, BER_Decoding_Error):
            # ASN1_Error may be raised by ASN1F_CHOICE
            return None, s

    def dissect(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> bytes
        try:
            return self._field.dissect(pkt, s)
        except (ASN1_Error, ASN1F_badsequence, BER_Decoding_Error):
            self._field.set_val(pkt, None)
            return s

    def build(self, pkt):
        # type: (ASN1_Packet) -> bytes
        if self._field.is_empty(pkt):
            return b""
        return self._field.build(pkt)

    def any2i(self, pkt, x):
        # type: (ASN1_Packet, Any) -> Any
        return self._field.any2i(pkt, x)

    def i2repr(self, pkt, x):
        # type: (ASN1_Packet, Any) -> str
        return self._field.i2repr(pkt, x)


class ASN1F_omit(ASN1F_field[None, None]):
    """
    ASN.1 field that is not specified. This is simply omitted on the network.
    This is different from ASN1F_NULL which has a network representation.
    """
    def m2i(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> Tuple[None, bytes]
        return None, s

    def i2m(self, pkt, x):
        # type: (ASN1_Packet, Optional[bytes]) -> bytes
        return b""


_CHOICE_T = Union['ASN1_Packet', Type[ASN1F_field[Any, Any]], 'ASN1F_PACKET']


class ASN1F_CHOICE(ASN1F_field[_CHOICE_T, ASN1_Object[Any]]):
    """
    Multiple types are allowed: ASN1_Packet, ASN1F_field and ASN1F_PACKET(),
    See layers/x509.py for examples.
    Other ASN1F_field instances than ASN1F_PACKET instances must not be used.
    """
    holds_packets = 1
    ASN1_tag = ASN1_Class_UNIVERSAL.ANY

    def __init__(self, name, default, *args, **kwargs):
        # type: (str, Any, *_CHOICE_T, **Any) -> None
        if "implicit_tag" in kwargs:
            err_msg = "ASN1F_CHOICE has been called with an implicit_tag"
            raise ASN1_Error(err_msg)
        self.implicit_tag = None
        for kwarg in ["context", "explicit_tag"]:
            setattr(self, kwarg, kwargs.get(kwarg))
        super(ASN1F_CHOICE, self).__init__(
            name, None, context=self.context,
            explicit_tag=self.explicit_tag
        )
        self.default = default
        self.current_choice = None
        self.choices = {}  # type: Dict[int, _CHOICE_T]
        self.pktchoices = {}
        for p in args:
            if hasattr(p, "ASN1_root"):
                p = cast('ASN1_Packet', p)
                # should be ASN1_Packet
                if hasattr(p.ASN1_root, "choices"):
                    root = cast(ASN1F_CHOICE, p.ASN1_root)
                    for k, v in root.choices.items():
                        # ASN1F_CHOICE recursion
                        self.choices[k] = v
                else:
                    self.choices[p.ASN1_root.network_tag] = p
            elif hasattr(p, "ASN1_tag"):
                if isinstance(p, type):
                    # should be ASN1F_field class
                    self.choices[int(p.ASN1_tag)] = p
                else:
                    # should be ASN1F_field instance
                    self.choices[p.network_tag] = p
                    self.pktchoices[hash(p.cls)] = (p.implicit_tag, p.explicit_tag)  # noqa: E501
            else:
                raise ASN1_Error("ASN1F_CHOICE: no tag found for one field")

    def m2i(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> Tuple[ASN1_Object[Any], bytes]
        """
        First we have to retrieve the appropriate choice.
        Then we extract the field/packet, according to this choice.
        """
        if len(s) == 0:
            raise ASN1_Error("ASN1F_CHOICE: got empty string")
        _, s = BER_tagging_dec(s, hidden_tag=self.ASN1_tag,
                               explicit_tag=self.explicit_tag)
        tag, _ = BER_id_dec(s)
        if tag in self.choices:
            choice = self.choices[tag]
        else:
            if self.flexible_tag:
                choice = ASN1F_field
            else:
                raise ASN1_Error(
                    "ASN1F_CHOICE: unexpected field in '%s' "
                    "(tag %s not in possible tags %s)" % (
                        self.name, tag, list(self.choices.keys())
                    )
                )
        if hasattr(choice, "ASN1_root"):
            # we don't want to import ASN1_Packet in this module...
            return self.extract_packet(choice, s, _underlayer=pkt)  # type: ignore
        elif isinstance(choice, type):
            return choice(self.name, b"").m2i(pkt, s)
        else:
            # XXX check properly if this is an ASN1F_PACKET
            return choice.m2i(pkt, s)

    def i2m(self, pkt, x):
        # type: (ASN1_Packet, Any) -> bytes
        if x is None:
            s = b""
        else:
            s = bytes(x)
            if hash(type(x)) in self.pktchoices:
                imp, exp = self.pktchoices[hash(type(x))]
                s = BER_tagging_enc(s,
                                    implicit_tag=imp,
                                    explicit_tag=exp)
        return BER_tagging_enc(s, explicit_tag=self.explicit_tag)

    def randval(self):
        # type: () -> RandChoice
        randchoices = []
        for p in self.choices.values():
            if hasattr(p, "ASN1_root"):
                # should be ASN1_Packet class
                randchoices.append(packet.fuzz(p()))  # type: ignore
            elif hasattr(p, "ASN1_tag"):
                if isinstance(p, type):
                    # should be (basic) ASN1F_field class
                    randchoices.append(p("dummy", None).randval())
                else:
                    # should be ASN1F_PACKET instance
                    randchoices.append(p.randval())
        return RandChoice(*randchoices)


class ASN1F_PACKET(ASN1F_field['ASN1_Packet', Optional['ASN1_Packet']]):
    holds_packets = 1

    def __init__(self,
                 name,  # type: str
                 default,  # type: Optional[ASN1_Packet]
                 cls,  # type: Type[ASN1_Packet]
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[int]
                 explicit_tag=None,  # type: Optional[int]
                 next_cls_cb=None,  # type: Optional[Callable[[ASN1_Packet], Type[ASN1_Packet]]]  # noqa: E501
                 ):
        # type: (...) -> None
        self.cls = cls
        self.next_cls_cb = next_cls_cb
        super(ASN1F_PACKET, self).__init__(
            name, None, context=context,
            implicit_tag=implicit_tag, explicit_tag=explicit_tag
        )
        if implicit_tag is None and explicit_tag is None and cls is not None:
            if cls.ASN1_root.ASN1_tag == ASN1_Class_UNIVERSAL.SEQUENCE:
                self.network_tag = 16 | 0x20  # 16 + CONSTRUCTED
        self.default = default

    def m2i(self, pkt, s):
        # type: (ASN1_Packet, bytes) -> Tuple[Any, bytes]
        if self.next_cls_cb:
            cls = self.next_cls_cb(pkt) or self.cls
        else:
            cls = self.cls
        if not hasattr(cls, "ASN1_root"):
            # A normal Packet (!= ASN1)
            return self.extract_packet(cls, s, _underlayer=pkt)
        diff_tag, s = BER_tagging_dec(s, hidden_tag=cls.ASN1_root.ASN1_tag,  # noqa: E501
                                      implicit_tag=self.implicit_tag,
                                      explicit_tag=self.explicit_tag,
                                      safe=self.flexible_tag,
                                      _fname=self.name)
        if diff_tag is not None:
            if self.implicit_tag is not None:
                self.implicit_tag = diff_tag
            elif self.explicit_tag is not None:
                self.explicit_tag = diff_tag
        if not s:
            return None, s
        return self.extract_packet(cls, s, _underlayer=pkt)

    def i2m(self,
            pkt,  # type: ASN1_Packet
            x  # type: Union[bytes, ASN1_Packet, None, ASN1_Object[Optional[ASN1_Packet]]]  # noqa: E501
            ):
        # type: (...) -> bytes
        if x is None:
            s = b""
        elif isinstance(x, bytes):
            s = x
        elif isinstance(x, ASN1_Object):
            if x.val:
                s = bytes(x.val)
            else:
                s = b""
        else:
            s = bytes(x)
            if not hasattr(x, "ASN1_root"):
                # A normal Packet (!= ASN1)
                return s
        return BER_tagging_enc(s,
                               implicit_tag=self.implicit_tag,
                               explicit_tag=self.explicit_tag)

    def any2i(self,
              pkt,  # type: ASN1_Packet
              x  # type: Union[bytes, ASN1_Packet, None, ASN1_Object[Optional[ASN1_Packet]]]  # noqa: E501
              ):
        # type: (...) -> 'ASN1_Packet'
        if hasattr(x, "add_underlayer"):
            x.add_underlayer(pkt)  # type: ignore
        return super(ASN1F_PACKET, self).any2i(pkt, x)

    def randval(self):  # type: ignore
        # type: () -> ASN1_Packet
        return packet.fuzz(self.cls())


class ASN1F_BIT_STRING_ENCAPS(ASN1F_BIT_STRING):
    """
    We may emulate simple string encapsulation with explicit_tag=0x04,
    but we need a specific class for bit strings because of unused bits, etc.
    """
    ASN1_tag = ASN1_Class_UNIVERSAL.BIT_STRING

    def __init__(self,
                 name,  # type: str
                 default,  # type: Optional[ASN1_Packet]
                 cls,  # type: Type[ASN1_Packet]
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[int]
                 explicit_tag=None,  # type: Optional[int]
                 ):
        # type: (...) -> None
        self.cls = cls
        super(ASN1F_BIT_STRING_ENCAPS, self).__init__(  # type: ignore
            name,
            default and bytes(default),
            context=context,
            implicit_tag=implicit_tag,
            explicit_tag=explicit_tag
        )

    def m2i(self, pkt, s):  # type: ignore
        # type: (ASN1_Packet, bytes) -> Tuple[Optional[ASN1_Packet], bytes]
        bit_string, remain = super(ASN1F_BIT_STRING_ENCAPS, self).m2i(pkt, s)
        if len(bit_string.val) % 8 != 0:
            raise BER_Decoding_Error("wrong bit string", remaining=s)
        if bit_string.val_readable:
            p, s = self.extract_packet(self.cls, bit_string.val_readable,
                                       _underlayer=pkt)
        else:
            return None, bit_string.val_readable
        if len(s) > 0:
            raise BER_Decoding_Error("unexpected remainder", remaining=s)
        return p, remain

    def i2m(self, pkt, x):  # type: ignore
        # type: (ASN1_Packet, Optional[ASN1_BIT_STRING]) -> bytes
        if not isinstance(x, ASN1_BIT_STRING):
            x = ASN1_BIT_STRING(
                b"" if x is None else bytes(x),  # type: ignore
                readable=True,
            )
        return super(ASN1F_BIT_STRING_ENCAPS, self).i2m(pkt, x)


class ASN1F_FLAGS(ASN1F_BIT_STRING):
    def __init__(self,
                 name,  # type: str
                 default,  # type: Optional[str]
                 mapping,  # type: List[str]
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[int]
                 explicit_tag=None,  # type: Optional[Any]
                 ):
        # type: (...) -> None
        self.mapping = mapping
        super(ASN1F_FLAGS, self).__init__(
            name, default,
            default_readable=False,
            context=context,
            implicit_tag=implicit_tag,
            explicit_tag=explicit_tag
        )

    def any2i(self, pkt, x):
        # type: (ASN1_Packet, Any) -> str
        if isinstance(x, str):
            if any(y not in ["0", "1"] for y in x):
                # resolve the flags
                value = ["0"] * len(self.mapping)
                for i in x.split("+"):
                    value[self.mapping.index(i)] = "1"
                x = "".join(value)
            x = ASN1_BIT_STRING(x)
        return super(ASN1F_FLAGS, self).any2i(pkt, x)

    def get_flags(self, pkt):
        # type: (ASN1_Packet) -> List[str]
        fbytes = getattr(pkt, self.name).val
        return [self.mapping[i] for i, positional in enumerate(fbytes)
                if positional == '1' and i < len(self.mapping)]

    def i2repr(self, pkt, x):
        # type: (ASN1_Packet, Any) -> str
        if x is not None:
            pretty_s = ", ".join(self.get_flags(pkt))
            return pretty_s + " " + repr(x)
        return repr(x)


class ASN1F_STRING_PacketField(ASN1F_STRING):
    """
    ASN1F_STRING that holds packets.
    """
    holds_packets = 1

    def i2m(self, pkt, val):
        # type: (ASN1_Packet, Any) -> bytes
        if hasattr(val, "ASN1_root"):
            val = ASN1_STRING(bytes(val))
        return super(ASN1F_STRING_PacketField, self).i2m(pkt, val)

    def any2i(self, pkt, x):
        # type: (ASN1_Packet, Any) -> Any
        if hasattr(x, "add_underlayer"):
            x.add_underlayer(pkt)
        return super(ASN1F_STRING_PacketField, self).any2i(pkt, x)


class ASN1F_STRING_ENCAPS(ASN1F_STRING_PacketField):
    """
    ASN1F_STRING that encapsulates a single ASN1 packet.
    """

    def __init__(self,
                 name,  # type: str
                 default,  # type: Optional[ASN1_Packet]
                 cls,  # type: Type[ASN1_Packet]
                 context=None,  # type: Optional[Any]
                 implicit_tag=None,  # type: Optional[int]
                 explicit_tag=None,  # type: Optional[int]
                 ):
        # type: (...) -> None
        self.cls = cls
        super(ASN1F_STRING_ENCAPS, self).__init__(
            name,
            default and bytes(default),  # type: ignore
            context=context,
            implicit_tag=implicit_tag,
            explicit_tag=explicit_tag
        )

    def m2i(self, pkt, s):  # type: ignore
        # type: (ASN1_Packet, bytes) -> Tuple[ASN1_Packet, bytes]
        val = super(ASN1F_STRING_ENCAPS, self).m2i(pkt, s)
        return self.cls(val[0].val, _underlayer=pkt), val[1]



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = python-can CANSocket
# scapy.contrib.status = loads

"""
Python-CAN CANSocket Wrapper.
"""

import time
import struct
import threading

from functools import reduce
from operator import add
from collections import deque

from scapy.config import conf
from scapy.supersocket import SuperSocket
from scapy.layers.can import CAN
from scapy.packet import Packet
from scapy.error import warning
from typing import (
    List,
    Type,
    Tuple,
    Dict,
    Any,
    Optional,
    cast,
)

from can import Message as can_Message
from can import CanError as can_CanError
from can import BusABC as can_BusABC
from can.interface import Bus as can_Bus

__all__ = ["CANSocket", "PythonCANSocket"]


class SocketMapper(object):
    """Internal Helper class to map a python-can bus object to
    a list of SocketWrapper instances
    """
    def __init__(self, bus, sockets):
        # type: (can_BusABC, List[SocketWrapper]) -> None
        """Initializes the SocketMapper helper class

        :param bus: A python-can Bus object
        :param sockets: A list of SocketWrapper objects which want to receive
                        messages from the provided python-can Bus object.
        """
        self.bus = bus
        self.sockets = sockets

    def mux(self):
        # type: () -> None
        """Multiplexer function. Tries to receive from its python-can bus
        object. If a message is received, this message gets forwarded to
        all receive queues of the SocketWrapper objects.
        """
        msgs = []
        while True:
            try:
                msg = self.bus.recv(timeout=0)
                if msg is None:
                    break
                else:
                    msgs.append(msg)
            except Exception as e:
                warning("[MUX] python-can exception caught: %s" % e)

        for sock in self.sockets:
            with sock.lock:
                for msg in msgs:
                    if sock._matches_filters(msg):
                        sock.rx_queue.append(msg)


class _SocketsPool(object):
    """Helper class to organize all SocketWrapper and SocketMapper objects"""
    def __init__(self):
        # type: () -> None
        self.pool = dict()  # type: Dict[str, SocketMapper]
        self.pool_mutex = threading.Lock()
        self.last_call = 0.0

    def internal_send(self, sender, msg):
        # type: (SocketWrapper, can_Message) -> None
        """Internal send function.

        A given SocketWrapper wants to send a CAN message. The python-can
        Bus object is obtained from an internal pool of SocketMapper objects.
        The given message is sent on the python-can Bus object and also
        inserted into the message queues of all other SocketWrapper objects
        which are connected to the same python-can bus object
        by the SocketMapper.

        :param sender: SocketWrapper which initiated a send of a CAN message
        :param msg: CAN message to be sent
        """
        if sender.name is None:
            raise TypeError("SocketWrapper.name should never be None")

        with self.pool_mutex:
            try:
                mapper = self.pool[sender.name]
                mapper.bus.send(msg)
                for sock in mapper.sockets:
                    if sock == sender:
                        continue
                    if not sock._matches_filters(msg):
                        continue

                    with sock.lock:
                        sock.rx_queue.append(msg)
            except KeyError:
                warning("[SND] Socket %s not found in pool" % sender.name)
            except can_CanError as e:
                warning("[SND] python-can exception caught: %s" % e)

    def multiplex_rx_packets(self):
        # type: () -> None
        """This calls the mux() function of all SocketMapper
        objects in this SocketPool
        """
        if time.monotonic() - self.last_call < 0.001:
            # Avoid starvation if multiple threads are doing selects, since
            # this object is singleton and all python-CAN sockets are using
            # the same instance and locking the same locks.
            return
        with self.pool_mutex:
            for t in self.pool.values():
                t.mux()
        self.last_call = time.monotonic()

    def register(self, socket, *args, **kwargs):
        # type: (SocketWrapper, Tuple[Any, ...], Dict[str, Any]) -> None
        """Registers a SocketWrapper object. Every SocketWrapper describes to
        a python-can bus object. This python-can bus object can only exist
        once. In case this object already exists in this SocketsPool, organized
        by a SocketMapper object, the new SocketWrapper is inserted in the
        list of subscribers of the SocketMapper. Otherwise a new python-can
        Bus object is created from the provided args and kwargs and inserted,
        encapsulated in a SocketMapper, into this SocketsPool.

        :param socket: SocketWrapper object which needs to be registered.
        :param args: Arguments for the python-can Bus object
        :param kwargs: Keyword arguments for the python-can Bus object
        """
        if "interface" in kwargs.keys():
            k = str(kwargs.get("interface", "unknown_interface")) + "_" + \
                str(kwargs.get("channel", "unknown_channel"))
        else:
            k = str(kwargs.get("bustype", "unknown_bustype")) + "_" + \
                str(kwargs.get("channel", "unknown_channel"))
        with self.pool_mutex:
            if k in self.pool:
                t = self.pool[k]
                t.sockets.append(socket)
                filters = [s.filters for s in t.sockets
                           if s.filters is not None]
                if filters:
                    t.bus.set_filters(reduce(add, filters))
                socket.name = k
            else:
                bus = can_Bus(*args, **kwargs)
                socket.name = k
                self.pool[k] = SocketMapper(bus, [socket])

    def unregister(self, socket):
        # type: (SocketWrapper) -> None
        """Unregisters a SocketWrapper from its subscription to a SocketMapper.

        If a SocketMapper doesn't have any subscribers, the python-can Bus
        get shutdown.

        :param socket: SocketWrapper to be unregistered
        """
        if socket.name is None:
            raise TypeError("SocketWrapper.name should never be None")

        with self.pool_mutex:
            try:
                t = self.pool[socket.name]
                t.sockets.remove(socket)
                if not t.sockets:
                    t.bus.shutdown()
                    del self.pool[socket.name]
            except KeyError:
                warning("Socket %s already removed from pool" % socket.name)


SocketsPool = _SocketsPool()


class SocketWrapper(can_BusABC):
    """Helper class to wrap a python-can Bus object as socket"""

    def __init__(self, *args, **kwargs):
        # type: (Tuple[Any, ...], Dict[str, Any]) -> None
        """Initializes a new python-can based socket, described by the provided
        arguments and keyword arguments. This SocketWrapper gets automatically
        registered in the SocketsPool.

        :param args: Arguments for the python-can Bus object
        :param kwargs: Keyword arguments for the python-can Bus object
        """
        super(SocketWrapper, self).__init__(*args, **kwargs)
        self.lock = threading.Lock()
        self.rx_queue = deque()  # type: deque[can_Message]
        self.name = None  # type: Optional[str]
        SocketsPool.register(self, *args, **kwargs)

    def _recv_internal(self, timeout):
        # type: (int) -> Tuple[Optional[can_Message], bool]
        """Internal blocking receive method,
        following the ``can_BusABC`` interface of python-can.

        This triggers the multiplex function of the general SocketsPool.

        :param timeout: Time to wait for a packet
        :return: Returns a tuple of either a can_Message or None and a bool to
                 indicate if filtering was already applied.
        """
        if not self.rx_queue:
            # Early return without locking if it looks like rx_queue is empty
            return None, True

        with self.lock:
            # It could be that 2 threads are using this same socket, so it's
            # necessary to check again if the queue was emptied between the
            # previous check and now
            if len(self.rx_queue) == 0:
                return None, True
            msg = self.rx_queue.popleft()
            return msg, True

    def send(self, msg, timeout=None):
        # type: (can_Message, Optional[int]) -> None
        """Send function, following the ``can_BusABC`` interface of python-can.

        :param msg: Message to be sent.
        :param timeout: Not used.
        """
        SocketsPool.internal_send(self, msg)

    def shutdown(self):
        # type: () -> None
        """Shutdown function, following the ``can_BusABC`` interface of
        python-can.
        """
        SocketsPool.unregister(self)
        super().shutdown()


class PythonCANSocket(SuperSocket):
    """Initializes a python-can bus object as Scapy PythonCANSocket.

    All provided keyword arguments, except *basecls* are forwarded to
    the python-can can_Bus init function. For further details on python-can
    check: https://python-can.readthedocs.io/

    Example:
        >>> socket = PythonCANSocket(bustype='socketcan', channel='vcan0', bitrate=250000)
    """  # noqa: E501
    desc = "read/write packets at a given CAN interface " \
           "using a python-can bus object"
    nonblocking_socket = True

    def __init__(self, **kwargs):
        # type: (Dict[str, Any]) -> None
        self.basecls = cast(Optional[Type[Packet]], kwargs.pop("basecls", CAN))
        self.can_iface = SocketWrapper(**kwargs)

    def recv_raw(self, x=0xffff):
        # type: (int) -> Tuple[Optional[Type[Packet]], Optional[bytes], Optional[float]]  # noqa: E501
        """Returns a tuple containing (cls, pkt_data, time)"""
        msg = self.can_iface.recv()

        hdr = msg.is_extended_id << 31 | msg.is_remote_frame << 30 | \
            msg.is_error_frame << 29 | msg.arbitration_id

        if conf.contribs['CAN']['swap-bytes']:
            hdr = struct.unpack("<I", struct.pack(">I", hdr))[0]

        dlc = msg.dlc << 24 | msg.is_fd << 18 | \
            msg.error_state_indicator << 17 | msg.bitrate_switch << 16
        pkt_data = struct.pack("!II", hdr, dlc) + bytes(msg.data)
        return self.basecls, pkt_data, msg.timestamp

    def send(self, x):
        # type: (Packet) -> int
        bx = bytes(x)
        msg = can_Message(is_remote_frame=x.flags == 0x2,
                          is_extended_id=x.flags == 0x4,
                          is_error_frame=x.flags == 0x1,
                          arbitration_id=x.identifier,
                          is_fd=bx[5] & 4 > 0,
                          error_state_indicator=bx[5] & 2 > 0,
                          bitrate_switch=bx[5] & 1 > 0,
                          dlc=x.length,
                          data=bx[8:])
        msg.timestamp = time.time()
        try:
            x.sent_time = msg.timestamp
        except AttributeError:
            pass
        self.can_iface.send(msg)
        return len(x)

    @staticmethod
    def select(sockets, remain=conf.recv_poll_rate):
        # type: (List[SuperSocket], Optional[float]) -> List[SuperSocket]
        """This function is called during sendrecv() routine to select
        the available sockets.

        :param sockets: an array of sockets that need to be selected
        :returns: an array of sockets that were selected and
            the function to be called next to get the packets (i.g. recv)
        """
        ready_sockets = \
            [s for s in sockets if isinstance(s, PythonCANSocket) and
             len(s.can_iface.rx_queue)]
        # checking the queue length without locking might sound
        # dangerous, but for the purpose of this select, if another
        # thread is reading the same socket, then even proper locking
        # wouldn't help
        if not ready_sockets:
            # yield this thread to avoid starvation
            time.sleep(0)

        SocketsPool.multiplex_rx_packets()
        return cast(List[SuperSocket], ready_sockets)

    def close(self):
        # type: () -> None
        """Closes this socket"""
        if self.closed:
            return
        super(PythonCANSocket, self).close()
        self.can_iface.shutdown()


CANSocket = PythonCANSocket



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Tabea Spahn <tabea.spahn@e-mundo.de>

# scapy.contrib.description = XCPScanner
# scapy.contrib.status = loads
import logging
from collections import namedtuple

from scapy.config import conf
from scapy.contrib.automotive import log_automotive
from scapy.contrib.automotive.xcp.cto_commands_master import \
    TransportLayerCmd, TransportLayerCmdGetSlaveId, Connect
from scapy.contrib.automotive.xcp.cto_commands_slave import \
    ConnectPositiveResponse, TransportLayerCmdGetSlaveIdResponse
from scapy.contrib.automotive.xcp.xcp import CTORequest, XCPOnCAN
from scapy.contrib.cansocket_native import CANSocket

# Typing imports
from typing import (
    Optional,
    List,
    Type,
    Iterator,
)

XCPScannerResult = namedtuple('XCPScannerResult', 'request_id response_id')


class XCPOnCANScanner:
    """
    Scans for XCP Slave on CAN
    """

    def __init__(self, can_socket, id_range=None,
                 sniff_time=0.1, add_padding=False, verbose=False):
        # type: (CANSocket, Optional[Iterator[int]], Optional[float], Optional[bool], Optional[bool]) -> None # noqa: E501

        """
        Constructor
        :param can_socket: Can Socket with XCPonCAN as basecls for scan
        :param id_range: CAN id range to scan
        :param sniff_time: time the scan waits for a response
                           after sending a request
        """

        conf.contribs["XCP"]["add_padding_for_can"] = add_padding
        self.__socket = can_socket
        self.id_range = id_range or range(0, 0x800)
        self.__sniff_time = sniff_time
        if verbose:
            log_automotive.setLevel(logging.DEBUG)

    def _scan(self, identifier, body, pid, answer_type):
        # type: (int, CTORequest, int, Type) -> List # noqa: E501

        log_automotive.info("Scan for id: " + str(identifier))
        flags = 'extended' if identifier >= 0x800 else 0
        cto_request = \
            XCPOnCAN(identifier=identifier, flags=flags) \
            / CTORequest(pid=pid) / body

        req_and_res_list, _unanswered = \
            self.__socket.sr(cto_request, timeout=self.__sniff_time,
                             verbose=False, multi=True)

        if len(req_and_res_list) == 0:
            log_automotive.info(
                "No answer for identifier: " + str(identifier))
            return []

        valid_req_and_res_list = filter(
            lambda req_and_res: req_and_res[1].haslayer(answer_type),
            req_and_res_list)
        return list(valid_req_and_res_list)

    def _send_connect(self, identifier):
        # type: (int) -> List[XCPScannerResult]
        """
        Sends CONNECT Message on the Control Area Network
        """
        all_slaves = []
        body = Connect(connection_mode=0x00)
        xcp_req_and_res_list = self._scan(identifier, body, 0xFF,
                                          ConnectPositiveResponse)

        for req_and_res in xcp_req_and_res_list:
            result = XCPScannerResult(response_id=req_and_res[1].identifier,
                                      request_id=identifier)
            all_slaves.append(result)
            log_automotive.info(
                "Detected XCP slave for broadcast identifier: " + str(
                    identifier) + "\nResponse: " + str(result))

        if len(all_slaves) == 0:
            log_automotive.info(
                "No XCP slave detected for identifier: " + str(identifier))
        return all_slaves

    def _send_get_slave_id(self, identifier):
        # type: (int) -> List[XCPScannerResult]
        """
        Sends GET_SLAVE_ID message on the Control Area Network
        """
        all_slaves = []
        body = TransportLayerCmd() / TransportLayerCmdGetSlaveId()
        xcp_req_and_res_list = \
            self._scan(
                identifier, body, 0xF2, TransportLayerCmdGetSlaveIdResponse)

        for req_and_res in xcp_req_and_res_list:
            response = req_and_res[1]
            # The protocol will also mark other XCP messages that might be
            # send as TransportLayerCmdGetSlaveIdResponse
            # -> Payload must be checked. It must include XCP
            if response.position_1 != 0x58 or response.position_2 != 0x43 or \
                    response.position_3 != 0x50:
                continue

            # Identifier that the master must use to send packets to the slave
            # and the slave will answer with
            request_id = \
                response["TransportLayerCmdGetSlaveIdResponse"].can_identifier

            result = XCPScannerResult(request_id=request_id,
                                      response_id=response.identifier)
            all_slaves.append(result)
            log_automotive.info(
                "Detected XCP slave for broadcast identifier: " + str(
                    identifier) + "\nResponse: " + str(result))

        return all_slaves

    def scan_with_get_slave_id(self):
        # type: () -> List[XCPScannerResult]
        """Starts the scan for XCP devices on CAN with the transport specific
        GetSlaveId Message"""
        log_automotive.info("Start scan with GetSlaveId id in range: " + str(
            self.id_range))

        for identifier in self.id_range:
            ids = self._send_get_slave_id(identifier)
            if len(ids) > 0:
                return ids

        return []

    def scan_with_connect(self):
        # type: () -> List[XCPScannerResult]
        log_automotive.info("Start scan with CONNECT id in range: " + str(
            self.id_range))
        results = []
        for identifier in self.id_range:
            result = self._send_connect(identifier)
            if len(result) > 0:
                results.extend(result)
        return results
